#!/usr/bin/env python
# -*- coding: utf-8 -*-

from lpu.common import logging
from lpu.common.logging import debug_print as dprint

if __name__ == '__main__':
    logger = logging.getColorLogger(__name__)
    print(logger.handlers)
    formatter = logger.handlers[0].formatter
    logger.info("-----")
    logger.info("testing colorized logger instance")
    logger.debug(logger.level)
    #logger.setLevel(logging.DEBUG)
    config = logging.using_config(logger, debug=True)
    logger.debug(logger.level)
    logger.debug("debug")
    try:
        logger.debug("debug", stack_info=True)
    except Exception as e:
        logger.exception(repr(e))
    logger.info("info")
    logger.warning("warning")
    logger.error("error")
    logger.critical("critical")
    try:
        raise RuntimeError("runtime error")
    except Exception as e:
        logger.exception(repr(e))
        formatter.setColor('exception', 'yellow')
        try:
            raise RuntimeError("runtime error 2")
        except Exception as e2:
            logger.exception(repr(e2))
            logger.exception(repr(e))
        logger.exception(e)
        logger.debug(e)

    logger.info("-----")
    logger.info("testing debug_print function")
    def f():
        dprint("debug print", 1)
        dprint("debug print", 2)
        dprint("debug print", 3)
    f()




# distutils: language=c++
# -*- coding: utf-8 -*-
# cython: profile=True

'''Configuration utility class for function settings'''

# C++ setting
from libcpp cimport bool

# Standard libraries
import json
from collections import OrderedDict
from collections import Iterable

# Local libraries
from lpu.common import logging
from lpu.common import validation
from lpu.common.logging import debug_print as dprint

cdef class ConfigData:
    '''Configuration data holder'''
    def __cinit__(self, _base=None, **args):
        if isinstance(_base, Config):
            self.__base = _base.data
        elif isinstance(_base, ConfigData):
            self.__base = _base
        elif isinstance(_base, dict):
            #self.__base = _base
            self.__base = dict2data(_base)
        else:
            self.__base = None
        self.__main = OrderedDict()
        if args:
            self.__main.update(args)

    def __contains__(self, key):
        cdef str first_key, remain_keys
        cdef object main = self.__main
        cdef object base = self.__base
        if isinstance(key, str) and key.find('.') >= 0:
            # chained key access
            try:
                first_key, remain_keys = key.split('.', 1)
                return self.__getitem__(first_key).__contains__(remain_keys)
            except:
                return False
        elif main.__contains__(key):
            return True
        elif base and base.__contains__(key):
            return True
        return False

    #def __getattr__(self, key):
    def __getattr__(self, str key):
        cdef str name
        try:
            return self.__getitem__(key)
        except:
            name = self.__class__.__name__
            raise AttributeError("'%s' object has no attribute '%s'" % (name, key))

    def __getitem__(self, key):
        cdef str msg
        cdef object main
        cdef object base
        cdef object value
        cdef str first_key, remain_keys
        main = self.__main
        if isinstance(key, str):
            if key.find('.') >= 0:
                # chained access key
                first_key, remain_keys = key.split('.', 1)
                return self.__getitem__(first_key).__getitem__(remain_keys)
            # first, check key existence in main dict object
            if key in main:
                return main[key]
            # othwerwise, check key existence in base (parent) dict object
            base = self.__base
            if base and key in base:
                # copy on reading
                value = base[key]
                if isinstance(value, ConfigData):
                    # derive, instead of copying
                    value = ConfigData(value)
                main[key] = value
                return value
                #return base[key]
            raise KeyError(key)
        elif isinstance(key, list):
            return [self[subkey] for subkey in key]
        elif isinstance(key, tuple):
            return tuple(self[subkey] for subkey in key)
        elif isinstance(key, Iterable):
            return (self[subkey] for subkey in key)
        else:
            msg = 'Invalid type of key object is given: {} (expected str or Iterable, but expected: {})'
            raise TypeError(msg.format(repr(key), type(key).__name__))

    def __iter__(self):
        cdef set s
        cdef list l
        cdef str key
        cdef object base = self.__base
        cdef object main = self.__main
        l = list(self.__main)
        s = set(l)
        if base:
            #s.update(base)
            for key in base:
                if key not in s:
                    s.add(key)
                    l.append(key)
        for key in l:
            if not key.startswith('_'):
                yield(key)

    def __len__(self):
        return len(set(self))

    def __repr__(self):
        #cdef str str_base
        cdef str str_params
        cdef str name = self.__class__.__name__
        cdef object main = self.__main
        cdef object base = self.__base
        #if base:
        #    str_base = repr(base)
        #else:
        #    str_base = ""
        str_params = get_key_val_str(main, False)
        if base:
            if str_params:
                return "{}({},{})".format(name, repr(base), str_params)
            else:
                return "{}({})".format(name, repr(base))
        else:
            if str_params:
                return "{}({})".format(name, str_params)
            else:
                return "{}()".format(name)
        #if str_params:
        #    return "%s(%r, %s)" % (name,self.__base,str_params)
        #else:
        #    return "%s(%s)" % (name,self.__base)

    def __setattr__(self, key, val):
        self.__setitem__(key, val)
        #if key.startswith('_'):
        #    raise KeyError('Key should not start with "_": %s' % key)
        #else:
        #    #self.__dict__.__setitem__(key, val)
        #    self.__main.__setitem__(key, val)

    def __setitem__(self, key, val):
        cdef str msg
        cdef object retrieved
        cdef ConfigData conf
        cdef object main = self.__main
        cdef object base = self.__base
        # check the key validity
        if not isinstance(key, str):
            raise TypeError('key value should be type of str, but given: %s' % type(key).__name__)
        elif key.startswith('_'):
            raise KeyError('key should not start with "_": %s' % key)
        # process for chained accessing
        if key.find('.') >= 0:
            # chained access key
            first_key, remain_keys = key.split('.', 1)
            if main.__contains__(first_key):
                retrieved = main.__getitem__(first_key)
                if isinstance(retrieved, ConfigData):
                    retrieved.__setitem__(remain_keys, val)
                else:
                    msg = "retrieved object with key '{}': {}, does not allow chained access with key '{}'"
                    raise KeyError(msg.format(first_key, repr(retrieved), remain_keys))
            elif base and base.__contains__(first_key):
                retrieved = self.__base.__getitem__(first_key)
                if isinstance(retrieved, ConfigData):
                    # deriving as base data
                    conf = ConfigData(retrieved)
                    main.__setitem__(first_key, conf)
                    conf.__setitem__(remain_keys, val)
                else:
                    msg = "retrieved object with key '{}': {}, does not allow chained access with key '{}'"
                    raise KeyError(msg.format(first_key, repr(retrieved), remain_keys))
            else:
                # setting new config data, and continue to chained access
                conf = ConfigData()
                main.__setitem__(first_key, conf)
                conf.__setitem__(remain_keys, val)
        else:
            if isinstance(val, dict):
                val = ConfigData(val)
            main.__setitem__(key, val)

cdef class Config:
    '''Configuration maintenance class'''

    def __cinit__(self, _base = None, **args):
        #self.data = ConfigData(_base)
        self.data = ConfigData(_base=_base)
        if args:
            self.update(args)

    def cast(self, name, typeof):
        val = self.require_any(name)
        if type(val) == typeof:
            return val
        else:
            casted = typeof(val)
            #self.data.__dict__[name] = newVal
            self.data[name] = casted
            return casted

    def has(self, key):
        if type(key) is str:
            return key in self
        elif isinstance(key, Iterable):
            return all(map(self.has, key))
        else:
            raise TypeError("Expected str or iterable type, but given: %s" % type(key).__name__)

    def get(self, key, default = None):
        if type(key) is str:
            if key in self:
                return self[key]
            else:
                return default
        elif isinstance(key, Iterable):
            return [self.get(elem, default) for elem in key]
        else:
            raise TypeError("Expected str or iterable type, but given: %s" % type(key).__name__)

    def items(self):
        for key in self:
            yield key, self[key]

    def load_json(self, str str_json, bool override=True):
        #self.update(json.loads(compat.to_str(strJSON)))
        #uniDict = json.loads(strJSON)
        d = json.loads(str_json, object_pairs_hook=OrderedDict)
        d = flat_dict(d, OrderedDict, True)
        #self.update(compat.to_str(uniDict))
        self.update(d, override)
        return self

    def require(self, name, desc = None, typeOf = None):
        val = self.require_any(name, desc)
        if typeOf:
            self.require_type(name, typeOf)
        return val

    def require_any(self, name, desc = None):
        if name not in self.data:
            if desc:
                raise KeyError('Configuration "%s" (%s) is not defined' % (name, desc))
            else:
                raise KeyError('Configuration "%s" is not defined' % (name))
        return self.data.__getitem__(name)

    def require_type(self, name, typeOf):
        val = self.require_any(name)
        if type(val) != typeOf:
            msg = 'Configuration "%s" should be type of %s, but given %s'
            logging.alert(msg % (name, typeOf, type(val)))
        return val

    def set(self, key, val):
        self[key] = val
        return self[key]

    def setdefault(self, key, val):
        if key not in self:
            return self.set(key, val)
        elif self[key] == None:
            return self.set(key, val)
        else:
            return self[key]

    def to_dict(self, key=None, ordered=False, upstream=False, recursive=True, flat=False):
        cdef type dtype
        cdef ConfigData data = self.data
        if ordered:
            dtype = OrderedDict
        else:
            dtype = dict
        if key:
            data = data[key]
        if flat:
            return flat_dict(data2dict(data, dtype, upstream, recursive), dtype, False)
        else:
            return data2dict(data, dtype, upstream, recursive)

    def to_json(self, key=None, upstream=False, **options):
        cdef object d
        d = self.to_dict(key, True, upstream)
        return json.dumps(d, **options)

    def update(self, _conf = None, _override=True, **args):
        if _conf:
            if _override:
                for key, val in _conf.items():
                    if val != None:
                        #dprint(key)
                        #dprint(val)
                        self[key] = val
            else:
                for key in _conf:
                    if key not in self:
                        self[key] = _conf[key]
        if args:
            self.update(args, _override)

    def __contains__(self, key):
        return self.data.__contains__(key)

    def __getitem__(self, key):
        return self.data.__getitem__(key)

    def __iter__(self):
        return self.data.__iter__()

    def __len__(self):
        return self.data.__len__()

    def __repr__(self):
        cdef type cls
        cdef str name
        cls = self.__class__
        name = cls.__name__
        #strParams = get_key_val_str(vars(self.data), False)
        return "%s(%r)" % (name, self.data)

    def __setitem__(self, key, val):
        self.data.__setitem__(key, val)

# type object is problematic in python 3.6?
#cdef object data2dict(object data, type dtype, bool upstream, bool recursive):
cdef object data2dict(object data, object dtype, bool upstream, bool recursive):
    #dprint("--")
    #dprint(data)
    #dprint(dtype)
    #dprint(upstream)
    #dprint(recursive)
    cdef ConfigData cdata
    if not isinstance(data, ConfigData):
        # as-is
        return data
    cdata = data
    if not recursive:
        if upstream:
            return dtype((key,data[key]) for key in data)
        else:
            return dtype((key,data[key]) for key in cdata.__main)
    else:
        if upstream:
            return dtype((key,data2dict(data[key],dtype,upstream,recursive)) for key in data)
        else:
            #dprint(cdata.__main)
            return dtype((key,data2dict(data[key],dtype,upstream,recursive)) for key in cdata.__main)

cdef object dict2data(object obj):
    cdef object key, value
    cdef ConfigData conf
    if not isinstance(obj, dict):
        # as-is
        return obj
    conf = ConfigData()
    for key, value in obj.items():
        conf[key] = dict2data(value)
    return conf

cdef get_key_val_str(object d, bool verbose):
    if verbose:
        items = ["%s=%r" % (t[0],t[1]) for t in d.items()]
    else:
        items = ["%s=%r" % (t[0],t[1]) for t in d.items() if not t[0].startswith('_')]
    return str.join(', ', items)

cdef list flat_items(object items, str prefix, bool chain_key):
    cdef object flatten = []
    cdef str str_prefix
    cdef str full_key
    if chain_key:
        if prefix:
            str_prefix = prefix + '.'
        else:
            str_prefix = ''
    else:
        str_prefix = ''
    for key, val in items:
        full_key = str_prefix + key
        if isinstance(val, dict):
            flatten += flat_items(val.items(), full_key, chain_key)
        else:
            flatten.append( (full_key,val) )
    return flatten
cdef object flat_dict(object d, type dtype, bool chain_key):
    cdef object flatten = dtype()
    for key, val in flat_items(d.items(), None, chain_key):
        if key not in flatten:
            flatten[key] = val
    return flatten




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import os
import sys

def defaultString(default):
    if default == "yes":
        return "[Y/n]"
    elif default == "no":
        return "[y/N]"
    else:
        return "[y/n]"

def getAnswer(default):
    ans = sys.stdin.readline().strip().lower()
    if ans in ("yes", "y"):
        return True
    elif ans in ("no", "n"):
        return False
    elif len(ans) == 0:
        return default
    else:
        return None

def askContinue(default=None):
    strYN = defaultString(default)
    sys.stderr.write("Do you want to continue? %s: " % strYN)
    sys.stderr.flush()
    ans = getAnswer(default)
    while ans == None:
        sys.stderr.write("Do you want to continue? %s: " % strYN)
        sys.stderr.flush()
        ans = getAnswer(default)
    if ans == False:
        sys.exit(1)

def askExistContinue(filepath, default=None):
    if os.path.exists(filepath):
        sys.stderr.write('"%s" is found. ' % filepath)
        askContinue(default)

def cmdPrompt(args):
    parser = argparse.ArgumentParser(description='Get absolute paths of given files or directories')
    parser.add_argument('--exist', '-e', type=str, help='asking whether continuing when specified file already exists')
    parser.add_argument('--continue', '-c', action='store_true', help='asking whether continuing')
    parser.add_argument('--yes', '-y', action='store_true', help='assign the default answer as "Yes"')
    parser.add_argument('--no', '-n', action='store_true', help='assign the default answer as "No"')
    parsed = parser.parse_args(args)
    default = None
    if parsed.yes:
        default = 'yes'
    if parsed.no:
        default = 'no'
    if getattr(parsed, 'continue'):
        return askContinue(default)
    elif parsed.exist:
        return askExistContinue(parsed.exist, default)
    else:
        sys.exit(1)

def main():
    cmdPrompt(sys.argv[1:])

if __name__ == '__main__':
    main()




# 3-rd party library
cimport numpy as np

# local library
from lpu.common.vocab cimport StringEnumerator

cdef class Vocab:
    cdef StringEnumerator src
    cdef StringEnumerator trg

    cdef void init(self)
    cpdef tuple ids_pair_to_str_pair(self, src_ids, trg_ids)
    cpdef list load_sent_pairs(self, str src_path, str trg_path)

cdef class Model:
    cdef Vocab vocab
    cdef np.ndarray trans_dist

    cdef void init(self)
    #cpdef double calc_pair_entropy(self, list src_sent, list trg_sent)
    cpdef double calc_pair_entropy(self, list src_sent, list trg_sent) except *
    cpdef double calc_entropy(self, list sent_pairs) except *
    cpdef void calc_and_save_scores(self, out_path, list sent_pairs)
    cpdef void save_align(self, out_path, threshold)

cdef class Trainer:
    cdef Model model
    cdef str src_path
    cdef str trg_path
    cdef list sent_pairs
    cdef np.ndarray cooc_src_trg

    cdef void init(self)
    #cpdef void init(self)
    #cdef inline void init(self):
    #    self.model = Model()

    cpdef np.ndarray calc_uniform_dist(self)
    cpdef np.ndarray count_cooccurrence(self, sent_pairs)
    cpdef void train(self, int iteration_limit)
    #cpdef void train(self, int iteration_limit) except *
    cpdef void train_first(self)
    cpdef void train_step(self)




# distutils: language=c++
# -*- coding: utf-8 -*-

'''Utilities for viewing I/O progress'''

# C++ setting
from libcpp cimport bool
from libcpp.string cimport string

# Standard libraries
from collections import Iterable
from datetime import datetime
#import io
import sys
import time

# Local libraries
from lpu.common import files
from lpu.common import logging
from lpu.common.colors import put_color
from lpu.common.logging import debug_print as dprint

logger = logging.getLogger(__name__)

from lpu.common cimport compat

if sys.version_info.major >= 3:
    bytes_to_str = compat.py3_bytes_to_str
else:
    bytes_to_str = compat.py2_bytes_to_str

cdef long DEFAULT_BUFFER_SIZE = 10 * (1024 ** 2) # 10MB

#cdef long BUFFER_SIZE = 4096
cdef str BACK_WHITE = '  \b\b'

#cdef double REFRESH = 1
cdef double REFRESH = 0.5

#cdef class ProgressCounter(object):
cdef class SpeedCounter(object):
    cdef readonly bool force
    cdef readonly str header
    cdef readonly double refresh
    cdef readonly double start_time, last_time
    cdef readonly long count, pos, last_count, max_count
    cdef readonly str color

    def __cinit__(self, str header="", long max_count=-1, double refresh=REFRESH, bool force=False, str color='green'):
        #logging.log("__CINIT__", color="cyan")
        self.refresh = refresh
        self.header = header 
        self.reset()
        self.force = force
        self.max_count = max_count
        self.color = color

    def add(self, unsigned long count=1, bool view=False):
        self.count += count
        if view:
            self.view()

    def flush(self):
        self.view(flush=True)

    def reset(self, refresh=None, header=None, force=None, color=None):
        cdef double now
        if self.last_time > self.start_time:
            self.flush()
            fobj = None
            if not self.force:
                if sys.stderr.isatty():
                    fobj = sys.stderr
                elif sys.stdout.isatty():
                    fobj = sys.stdout
            else:
                fobj = sys.stderr
            if fobj:
                fobj.write("\n")
        now = time.time()
        self.start_time = now
        self.last_time  = now
        self.count = 0
        self.last_count = 0
        self.pos = 0
        if refresh != None:
            self.refresh = refresh
        if header != None:
            self.header = header
        if force != None:
            self.force = force
        if color != None:
            self.color = color

    def set_count(self, unsigned long count, bool view=False):
        self.count = count
        if view:
            self.view()

    def set_position(self, unsigned long position, bool view=False):
        self.pos = position
        if view:
            self.view()

    def view(self, bool flush=False):
        cdef double now, delta_time, delta_count
        cdef str str_elapsed, str_rate, str_ratio
        cdef str str_timestamp, str_header, str_about
        cdef str str_print
        cdef bool show_bytes
        now = time.time()
        delta_time  = now - self.last_time
        if not flush:
            if delta_time < self.refresh:
                return False
        fobj = None
        if not self.force:
            if sys.stderr.isatty():
                fobj = sys.stderr
            elif sys.stdout.isatty():
                fobj = sys.stdout
        else:
            fobj = sys.stderr
        if fobj:
            delta_count = self.count - self.last_count
            show_bytes = False
            if self.count == self.pos:
                # bytes mode
                show_bytes = True
            str_rate = about(delta_count / delta_time, show_bytes)
            if self.header:
                str_header = "%s: " % self.header
            else:
                str_header = ""
            if self.max_count > 0:
                if self.pos > 0:
                    str_ratio = "(%.2f%%) " % (self.pos * 100.0 / self.max_count)
                else:
                    str_ratio = "(%.2f%%) " % (self.count * 100.0 / self.max_count)
            else:
                str_ratio = ""
            try:
                #strTimeStamp = datetime.datetime.now().strftime("%Y/%m/%d %H:%M:%S")
                #str_timestamp = datetime.now().strftime("%Y/%m/%d %H:%M:%S")
                str_timestamp = datetime.now().strftime("%Y/%m/%d %H:%M:%S")
                fobj.write("\r")
                str_elapsed = format_time(now - self.start_time)
                #logging.debug(elapsed)
                #fobj.write("%s%s %s%s [%s/s] [%s]" % (showName, about(self.count,show_bytes), strRatio, strElapsed, strRate, strTimeStamp))
                #fobj.write("[%s] %s%s %s%s [%s/s]" % (strTimeStamp, showName, about(self.count,show_bytes), strRatio, strElapsed, strRate))
                #fobj.write("  \b\b")
                str_about = about(self.count, show_bytes)
                str_print = "[%s] %s%s %s%s [%s/s]%s" % (str_timestamp, str_header, str_about, str_ratio, str_elapsed, str_rate, BACK_WHITE)
                #logging.put_color(str_print, self.color, newline=False)
                str_print = put_color(str_print, self.color)
                #logger.debug(str_print)
                fobj.write(str_print)
            except Exception as e:
                #dprint(str_print)
                #dprint(type(str_print))
                #dprint(self.color)
                #dprint(type(self.color))
                logger.exception(e)
        #if fobj and flush:
        #    fobj.write("\n")
        self.last_time  = now
        self.last_count = self.count
        return True

    def __del__(self):
        self.reset()

    def __enter__(self):
        #logger.debug("__enter__")
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        #logger.debug("__exit__")
        self.reset()

#cdef class ProgressReader(object):
cdef class FileReader(object):
    #cdef ProgressCounter counter
    cdef SpeedCounter counter
    cdef object source

    def __cinit__(self, source, str header="", double refresh=REFRESH, bool force=False):
    #def __cinit__(self, source, header="", double refresh=REFRESH, bool force=False):
        if isinstance(source, str):
            #self.source = files.open(source, 'r')
            #self.source = files.open(source, 'rb')
            if not header:
                header = "reading file '%s'" % source
            #self.source = files.open(source, 'rt')
            self.source = files.open(source, 'rb')
        #elif isinstance(source, io.IOBase):
        elif isinstance(source, files.FileType):
            self.source = source
        else:
            raise TypeError("FileReader() expected iterable str or file type, but given %s found" % type(source).__name__)
        size = files.rawsize(self.source)
        #self.counter = ProgressCounter(header=header, refresh=refresh, force=force, max_count=size)
        self.counter = SpeedCounter(header=header, max_count=size, refresh=refresh, force=force)

    #def __del__(self):
    #    print("DEL")
    #    self.close()

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.source:
            #self.counter.flush()
            self.counter.reset()
            self.counter = None
            self.source.close()
            self.source = None

    cpdef bytes read(self, size):
        if self.source:
            buf = self.source.read(size)
            self.counter.add(len(buf))
            try:
                #self.counter.set_position(files.rawtell(self.source))
                self.counter.set_position(self.tell())
            except Exception as e:
                #logger.debug(e)
                pass
            self.counter.view()
            return buf

    def read_byte_chunks(self, bs = DEFAULT_BUFFER_SIZE):
        cdef bytes buf
        while True:
            buf = self.read(bs)
            if not buf:
                break
            yield buf
        self.close()

    cdef bytes _read_byte_line(self, bool countup=False):
        cdef bytes line
        if self.source:
            line = self.source.readline()
            if countup:
                self.counter.add(len(line))
            try:
                #self.counter.set_position(files.rawtell(self.source))
                self.counter.set_position(self.tell())
            except Exception as e:
                pass
            self.counter.view()
            return line
    cpdef bytes read_byte_line(self):
        return self._read_byte_line(True)

    def read_byte_lines(self):
        cdef bytes line
        while True:
            line = self.read_byte_line()
            if not line:
                break
            yield line
        self.close()

    cpdef str readline(self):
        cdef bytes line
        line = self._read_byte_line(False)
        if line:
            self.counter.add(1)
            return bytes_to_str(line)
        return None

    cpdef long tell(self) except *:
        return files.rawtell(self.source)

    def __iter__(self):
        cdef str line
        while True:
            line = self.readline()
            if not line:
                break
            yield line
        self.close()

    def __enter__(self):
        #logger.debug("__enter__")
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        #logger.debug("__exit__")
        self.close()

cdef class Iterator(object):
    cdef SpeedCounter counter
    cdef object source

    def __cinit__(self, source, str header="", double refresh=REFRESH, bool force=False, long max_count=-1):
        if isinstance(source, Iterable):
            self.source = source
        else:
            raise TypeError("Iterator() expected iterable type, but %s found" % type(source).__name__)
        self.counter = SpeedCounter(header=header, max_count=max_count, refresh=refresh, force=force)

    def __dealloc__(self):
        self.close()

    cdef close(self):
        if self.source is not None:
            #self.counter.flush()
            self.counter.reset()
            self.counter = None
            self.source = None

    def __iter__(self):
        cdef object obj
        if self.source is not None:
            for obj in self.source:
                self.counter.add(1, view=True)
                yield obj
        self.close()

    def __len__(self):
        return len(self.source)

    def __enter__(self):
        #logger.debug("__enter__")
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        #logger.debug("__exit__")
        self.close()

cdef format_time(seconds):
    cdef unsigned char show_seconds, show_minutes
    cdef unsigned long show_hours
    seconds = int(seconds)
    show_seconds = seconds % 60
    show_minutes = (seconds / 60) % 60
    show_hours = seconds / (60*60)
    return "%02d:%02d:%02d" % (show_hours,show_minutes,show_seconds)

def about(num, bool show_bytes = False):
    if show_bytes:
        if num >= 2 ** 30:
            show = num / float(2 ** 30)
            return "%.3fGiB" % show
        elif num >= 2 ** 20:
            show = num / float(2 ** 20)
            return "%.3fMiB" % show
        elif num >= 2 ** 10:
            show = num / float(2 ** 10)
            return "%.3fKiB" % show
        else:
            return "%.3f" % num
    else:
        if num >= 10 ** 9:
            show = num / float(10 ** 9)
            return "%.3fG" % show
        elif num >= 10 ** 6:
            show = num / float(10 ** 6)
            return "%.3fM" % show
        elif num >= 10 ** 3:
            show = num / float(10 ** 3)
            return "%.3fk" % show
        else:
            return "%.3f" % num

cpdef open(path, str header=""):
    return FileReader(path, header)

cpdef pipe_view(filepaths, mode='bytes', str header=None, refresh=REFRESH, outfunc=None):
    #cdef str strBuf
    cdef bytes buf
    cdef long max_count = -1
    cdef long delta = 1
    cdef SpeedCounter counter
    if refresh < 0:
        refresh = REFRESH
    infiles = [files.open(fpath, 'rb') for fpath in filepaths]
    if infiles:
        #if mode == 'bytes':
        try:
            max_count = sum(map(files.rawsize, infiles))
        except Exception as e:
            logger.debug(e)
            max_count = 0
    else:
        infiles = [files.bin_stdin]
    counter = SpeedCounter(header=header, refresh=refresh, max_count=max_count)
    for infile in infiles:
        while True:
            buf = infile.read(DEFAULT_BUFFER_SIZE)
            if mode == 'bytes':
                delta = len(buf)
            elif mode == 'lines':
                delta = buf.count(b"\n")
            if not buf:
                break
            #counter.add(view=True)
            if not outfunc:
                files.bin_stdout.write(buf)
            counter.add(delta)
            counter.set_position(counter.pos + len(buf))
            counter.view()
    #counter.flush()
    counter.reset()

cpdef view(source, str header = None, long max_count = -1, env = True):
    if env and logging.get_quiet_status():
        # as-is (without progress view)
        return source
    elif isinstance(source, (FileReader,Iterator)):
        return source
    #elif isinstance(source, (str,io.IOBase)):
    elif isinstance(source, (str,bytes,files.FileType)):
        if not header:
            #header = "reading file"
            header = "reading file '{}'".format(source)
        return FileReader(source, header)
        #return FileReader(source, header, force=True)
    elif isinstance(source, Iterable):
        if not header:
            header = "iterating"
        if max_count < 0:
            if hasattr(source, '__len__'):
                max_count = len(source)
        return Iterator(source, header, max_count=max_count)
        #return Iterator(source, header, max_count=max_count, force=True)
    else:
        raise TypeError("view() expected file or iterable type, but %s found" % type(source).__name__)




# distutils: language=c++
# -*- coding: utf-8 -*-

'''Utility functions for Python 2/3 compatibility'''

import sys
#import collections
import types

cdef bytes py2_bytes_to_str(bytes b):
    # as-is
    return b
cdef str py3_bytes_to_str(bytes b):
    return str(b, 'utf-8', errors='backslashreplace')

cdef bytes __py2__to_bytes(s):
    '''
    convert to byte string
    '''
    if isinstance(s, unicode):
        return s.encode('utf-8')
    else:
        return bytes(s)
cdef bytes __py3__to_bytes(s):
    '''
    convert to byte string
    '''
    if isinstance(s, str):
        return bytes(s, 'utf-8')
    elif isinstance(s, bytes):
        return s
    else:
        return bytes(str(s), 'utf-8')

cdef str __py2__to_str(data):
    '''
    convert to object based on standard strings
    '''
    if isinstance(data, basestring):
        if isinstance(data, unicode):
            return data.encode('utf-8')
        else:
            return data
    #elif isinstance(data, collections.Mapping):
    #    return type(data)(map(to_str, data.iteritems()))
    #elif isinstance(data, collections.Iterable):
    #    return type(data)(map(to_str, data))
    else:
        return str(data)
cdef str __py3__to_str(data):
    '''
    convert to object based on standard strings
    '''
    if isinstance(data, str):
        return data
    elif isinstance(data, bytes):
        return str(data, 'utf-8')
    #elif isinstance(data, collections.Mapping):
    #    return type(data)(map(to_str, data.items()))
    #elif isinstance(data, collections.Iterable):
    #    return type(data)(map(to_str, data))
    else:
        return str(data)

cpdef unicode __py2__to_unicode(s):
    '''
    convert to unicode string
    '''
    return unicode(s, 'utf-8')
cpdef unicode __py3__to_unicode(s):
    '''
    convert to unicode string
    '''
    if isinstance(s, bytes):
        return str(s, 'utf-8')
    else:
        return str(s)

cpdef __py3__MethodType(function, instance, cls=None):
    # cls (class type) is not used
    return types.MethodType(function, instance)

if sys.version_info.major == 2:
    # Python2
    import itertools
    to_bytes   = __py2__to_bytes
    to_str     = __py2__to_str
    to_unicode = __py2__to_unicode
    range = xrange
    reduce = reduce
    zip   = itertools.izip
    MethodType = types.MethodType
elif sys.version_info.major == 3:
    # Python3
    import functools
    to_bytes   = __py3__to_bytes
    to_str     = __py3__to_str
    to_unicode = __py3__to_unicode
    range = range
    reduce = functools.reduce
    zip   = zip
    MethodType = __py3__MethodType
else:
    raise SystemError("Unsupported python version: %s" % sys.version)




# distutils: language=c++
# -*- coding: utf-8 -*-

'''classes handling rule table records'''

# Standard libraries
import math

# Local libraries
from lpu.common import compat
from lpu.common import logging
from lpu.common import numbers

# Matching Method
#   tree:    Full Match (Tree Match)
#   tag:     Tag Match (Phrase Structure Match)
#   string:  Symbol String Match (Superficial Match)
MATCH=['tree', 'tag', 'string', 'ftree']

cdef class CoOccurrence:
    cdef public float src, trg, cooc
    def __cinit__(self, float src=0.0, float trg=0.0, float cooc=0):
        self.src  = src
        self.trg  = trg
        self.cooc = cooc

    # calculate P(e|f) (src -> trg)
    cpdef float calcEGFP(self):
        return self.cooc / self.src
    property egfp:
        def __get__(self): return self.calcEGFP()

    # calculate P(f|e) (trg -> src)
    cpdef float calcFGEP(self):
        return self.cooc / self.trg
    property fgep:
        def __get__(self): return self.calcFGEP()

    cpdef CoOccurrence getReversed(self):
        return CoOccurrence(self.trg, self.src, self.cooc)

    cdef void round(self, digits=6):
        self.src  = round(self.src, digits)
        self.trg  = round(self.trg, digits)
        self.cooc = round(self.cooc, digits)

    cpdef void setCounts(self, object src=None, object trg=None, object cooc=None):
        if src:
            self.src = src
        if trg:
            self.trg = trg
        if cooc:
            self.cooc = cooc

    cpdef str to_str(self, float margin = 0):
        cdef str name = self.__class__.__name__
        cdef str mod  = self.__class__.__module__
        cdef str src  = str(numbers.toNumber(self.src,margin))
        cdef str trg  = str(numbers.toNumber(self.trg,margin))
        cdef str cooc = str(numbers.toNumber(self.cooc,margin))
        return "%s.%s(src = %s, trg = %s, cooc = %s)" % (mod, name, src, trg, cooc)

    def __str__(self):
        return self.to_str(0)

cdef class Record(object):
    cdef public str src, trg
    #cdef public long src, trg
    cdef public dict features
    cdef public CoOccurrence counts
    cdef public set aligns
    cdef public object data

    def __cinit__(self):
      self.src = ""
      self.trg = ""
      #self.src = ()
      #self.trg = ()
      #self.src = 0
      #self.trg = 0
      self.features = {}
      self.counts = CoOccurrence()
#      self.aligns = []
      self.aligns = set()

    cpdef fixOrder(self):
        return False

    def getSrcSymbols(self):
        return self.getSymbols(self.src)
    src_symbols = property(getSrcSymbols)

    def getSrcTerms(self):
        return self.getTerms(self.src)
    srcTerms = property(getSrcTerms)

    def getTrgSymbols(self):
        return self.getSymbols(self.trg)
    trgSymbols = property(getTrgSymbols)

    def getTrgTerms(self):
        return self.getTerms(self.trg)
    trgTerms = property(getTrgTerms)

    cpdef getAlignMap(self):
        return getAlignMap(self.aligns, reverse = False)
    #alignMap = property(getAlignMap)
    property alignMap:
        def __get__(self): return self.getAlignMap()

    cpdef getAlignMapRev(self):
        return getAlignMap(self.aligns, reverse = True)
    #alignMapRev = property(getAlignMapRev)
    property alignMapRev:
        def __get__(self): return self.getAlignMapRev()

    cpdef Record getReversed(self):
        cdef Record recRev
        cdef dict revFeatures
        recRev = self.__class__()
        recRev.src = self.trg
        recRev.trg = self.src
        recRev.counts = self.counts.getReversed()
#        debug.log(self.to_str())
#        debug.log(self.aligns)
#        recRev.aligns = getRevAligns(self.aligns)
        recRev.aligns = getRevAlignSet(self.aligns)
        revFeatures = {}
        if 'egfp' in self.features:
            revFeatures[intern('fgep')] = self.features['egfp']
        if 'egfl' in self.features:
            revFeatures[intern('fgel')] = self.features['egfl']
        if 'fgep' in self.features:
            revFeatures[intern('egfp')] = self.features['fgep']
        if 'fgel' in self.features:
            revFeatures[intern('egfl')] = self.features['fgel']
        if 'p' in self.features:
            revFeatures[intern('p')] = self.features['p']
        revFeatures[intern('w')] = len(self.srcTerms)
        recRev.features = revFeatures
        return recRev

    @staticmethod
    def getTerms(phrase):
        if isinstance(phrase, str):
            phrase = phrase.split(' ')
        return phrase

    @staticmethod
    def getSymbols(phrase):
        if isinstance(phrase, str):
            phrase = phrase.split(' ')
        return phrase

class MosesRecord(Record):
    def __init__(self, line = "", delim = '|||'):
        Record.__init__(self)
        self.delim = delim
        self.loadLine(line, delim)

    def loadLine(self, line, delim = '|||'):
        if line:
            fields = line.strip().split(delim)
            self.src = intern( fields[0].strip() )
            self.trg = intern( fields[1].strip() )
            self.features = getMosesFeatures(fields[2])
#            self.aligns = fields[3].strip().split()
            self.aligns = getAlignSet( fields[3] )
            listCounts = getCounts(fields[4])
            self.counts.setCounts(trg = listCounts[0], src = listCounts[1], co = listCounts[2])

#    def getSrcSymbols(self):
#        return self.src.split(' ')
#    src_symbols = property(getSrcSymbols)
#
#    def getSrcTerms(self):
#        return self.src.split(' ')
#    srcTerms = property(getSrcTerms)
#
#    def getTrgSymbols(self):
#        return self.trg.split(' ')
#    trgSymbols = property(getTrgSymbols)
#
#    def getTrgTerms(self):
#        return self.trg.split(' ')
#    trgTerms = property(getTrgTerms)

    def to_str(self, s = ' ||| '):
        strFeatures = getStrMosesFeatures(self.features)
#        strAligns = str.join(' ', self.aligns)
        strAligns = str.join(' ', sorted(self.aligns) )
        self.counts.round()
        strCounts   = "%s %s %s" % (self.counts.trg, self.counts.src, self.counts.cooc)
        #buf = str.join(s, [self.src, self.trg, strFeatures, strAligns, strCounts]) + "\n"
        buf = str.join(s, [self.src, self.trg, strFeatures, strAligns, strCounts])
#        buf = str.join(s, [str(self.src), str(self.trg), strFeatures, strAligns, strCounts]) + "\n"
        return buf

def getAlignMap(aligns, reverse = False):
    alignMap = {}
    for align in aligns:
        (s, t) = map(int, align.split('-'))
        if reverse:
            alignMap.setdefault(t, []).append(s)
        else:
            alignMap.setdefault(s, []).append(t)
    return alignMap

def getAlignSet(strField):
    return set(strField.strip().split())

def getCounts(field):
#    return map(getNumber, field.split())
#    return map(numbers.toNumber, field.split())
    return list(map(numbers.toNumber, field.split()))

def getNumber(anyNum, margin = 0):
    numFloat = float(anyNum)
    numInt = int(round(numFloat))
    if margin > 0:
        if abs(numInt - numFloat) < margin:
            return numInt
        else:
            return numFloat
    elif numFloat == numInt:
        return numInt
    else:
        return numFloat

#def getRevAligns(aligns):
def getRevAlignSet(aligns):
#    revAlignList = []
    revAlignSet = set()
#    debug.log(aligns)
    for a in aligns:
      (s, t) = map(int, a.split('-'))
#      revAlignList.append( "%d-%d" % (t, s) )
      revAlignSet.add( "%d-%d" % (t, s) )
#    return sorted(revAlignList)
    return sorted(revAlignSet)


def getMosesFeatures(field):
    features = {}
    scores = map(getNumber, field.split())
    features[intern('fgep')] = scores[0]
    features[intern('fgel')] = scores[1]
    features[intern('egfp')] = scores[2]
    features[intern('egfl')] = scores[3]
    return features

def getStrMosesFeatures(dicFeatures):
    '''convert back the feature dictionary to score string separated by space'''
    scores = []
    scores.append( dicFeatures.get('fgep', 0) )
    scores.append( dicFeatures.get('fgel', 0) )
    scores.append( dicFeatures.get('egfp', 0) )
    scores.append( dicFeatures.get('egfl', 0) )
    return str.join(' ', map(str,scores))


cdef class TravatarRecord(Record):
    cdef str delim

    def __cinit__(self, object line="", str delim='|||'):
        Record.__init__(self)
        self.delim = delim
        self.loadLine(compat.to_str(line), delim)

#    cpdef getSrcSymbols(self):
#      return getTravatarSymbols(self.src)
#    #src_symbols = property(getSrcSymbols)
#    property src_symbols:
#        def __get__(self): return self.getSrcSymbols()
#
#    cpdef getSrcTerms(self):
#      return getTravatarTerms(self.src)
#    #srcTerms = property(getSrcTerms)
#    property srcTerms:
#        def __get__(self): return self.getSrcTerms()
#
#    cpdef getTrgSymbols(self):
#      return getTravatarSymbols(self.trg)
#    #trgSymbols = property(getTrgSymbols)
#    property trgSymbols:
#        def __get__(self): return self.trgSrcSymbols()
#
#    cpdef getTrgTerms(self):
#      return getTravatarTerms(self.trg)
#    #trgTerms = property(getTrgTerms)
#    property trgTerms:
#        def __get__(self): return self.trgTerms()

    cpdef fixOrder(self):
        cdef str src, trg
        cdef object changed
        src, trg = TravatarRecord.fixOrderStrings(self.src, self.trg)
        changed = bool(src != self.src)
        self.src = src
        self.trg = trg
        return changed

    cpdef TravatarRecord getReversed(self):
      cdef TravatarRecord recRev = TravatarRecord()
      cdef dict revFeatures
      recRev.src = self.trg
      recRev.trg = self.src
      recRev.counts = self.counts.getReversed()
  #    recRev.aligns = record.getRevAligns(self.aligns)
      recRev.aligns = getRevAlignSet(self.aligns)
      revFeatures = {}
      if 'egfp' in self.features:
        revFeatures['fgep'] = self.features['egfp']
      if 'egfl' in self.features:
        revFeatures['fgel'] = self.features['egfl']
      if 'fgep' in self.features:
        revFeatures['egfp'] = self.features['fgep']
      if 'fgel' in self.features:
        revFeatures['egfl'] = self.features['fgel']
      if 'p' in self.features:
        revFeatures['p'] = self.features['p']
      revFeatures['w'] = len(self.srcTerms)
      recRev.features = revFeatures
      return recRev

#    cpdef loadLine(self, str line, str delim = '|||'):
    #def loadLine(self, line, delim = '|||'):
    cpdef loadLine(self, str line, str delim = '|||'):
        cdef list fields
        cdef list listCounts
        if line:
            fields = line.strip().split(delim)
            self.src = fields[0].strip()
            self.trg = fields[1].strip()
            #self.src = vocab.phrase2id(fields[0])
            #self.trg = vocab.phrase2id(fields[1])
            self.features = getTravatarFeatures(fields[2])
            listCounts = getCounts(fields[3])
            if len(listCounts) == 3:
                self.counts.setCounts(cooc = listCounts[0], src = listCounts[1], trg = listCounts[2])
            #self.aligns = fields[4].strip().split()
            self.aligns = getAlignSet( fields[4] )
            #self.data = pd.DataFrame()
            #self.data = pd.Series()

    cpdef syncTags(self):
        self.sync = False
        src_symbols = self.src.split(' ')
        trgSymbols = self.trg.split(' ')
#        if len(src_symbols) >= 4 and src_symbols[0][0] != '"':
        if isTree(src_symbols):
            # src is tree (s-expression)
            if len(trgSymbols) >= 2 and trgSymbols[-2] == "@":
                # trg is tagged
                pass
            else:
                # trg is not tagged
                srcTag = src_symbols[0]
                srcNonTerminals = []
                for s in src_symbols:
                    if s[0:1] == 'x' and s[1:2].isdigit():
                        srcNonTerminals.append(s)
#                        if len(s) > 3:
#                            srcNonTerminals.append(s[3:])
#                        else:
#                            srcNonTerminals.append('X')
                for i, s in enumerate(trgSymbols):
#                    print("TRG SYMBOLS: %s" % s)
                    if s[0:1] == 'x' and s[1:2].isdigit():
#                        print("NONTERMINAL!")
                        if len(s) > 3:
                            # symbol is tagged
                            pass
                        else:
                            # symbol is not tagged
                            index = int(s[1])
                            if self.sync:
                                trgSymbols[i] = srcNonTerminals[index]
                            else:
                                trgSymbols[i] = s + ":X"
#                self.trg = intern(self.trg + " @ " + srcTag)
                trgSymbols.append("@")
                if self.sync:
                    trgSymbols.append(srcTag)
                else:
                    trgSymbols.append('X')
                self.trg = intern(str.join(' ', trgSymbols))
        if len(src_symbols) == 3:
            tag = src_symbols[2]
            if src_symbols[0] == "x0:"+tag:
                # src is unary cycle
                #self.src = ""
                self.src = ()

#    cpdef toDict(self):
#        d = {}
#        d['src'] = self.src
#        d['trg'] = self.trg
#        d['features'] = self.features
#        d['counts'] = self.counts
#        d['aligns'] = self.aligns
#        return d

    cpdef to_str(self, s = ' ||| '):
      strFeatures = getStrTravatarFeatures(self.features)
      self.counts.round()
      strCounts = ""
      if self.counts.cooc > 0:
          #self.counts.cooc = round(self.counts.cooc, 6)
          #strCounts = "%s %s %s" % (self.counts.cooc, self.counts.src, self.counts.trg)
          strCounts = "%s %s %s" % (round(self.counts.cooc,6), round(self.counts.src,6), round(self.counts.trg,6))
  #    strAligns = str.join(' ', self.aligns)
      strAligns = str.join(' ', sorted(self.aligns))
      #buf = str.join(s, [self.src, self.trg, strFeatures, strCounts, strAligns]) + "\n"
      buf = str.join(s, [self.src, self.trg, strFeatures, strCounts, strAligns])
      return buf

    @staticmethod
    def fixOrderStrings(str src, str trg):
        cdef list srcElements = []
        cdef list trgElements = []
        cdef dict replaceDict = {}
        cdef str newSymbol
        cdef object changed = False
        for s in src.split(' '):
            if s[0] == 'x' and s[1].isdigit():
                replaceDict[s[1]] = str(len(replaceDict))
                newSymbol = 'x'+replaceDict[s[1]]+s[2:]
                srcElements.append(newSymbol)
                if s[1] != newSymbol[1]:
                    changed = True
            else:
                # as-is
                srcElements.append(s)
        if not changed:
            return (src, trg)
        for s in trg.split(' '):
            if s[0] == 'x' and s[1].isdigit():
                newSymbol = 'x'+replaceDict[s[1]]+s[2:]
                trgElements.append(newSymbol)
            else:
                # as-is
                trgElements.append(s)
        return (str.join(' ',srcElements), str.join(' ',trgElements))

    @staticmethod
    def getSymbols(rule, hiero=False):
        symbols = []
        for s in rule.split(' '):
          if len(s) < 2:
            if s == "@":
              break
          elif s[0] == '"' and s[-1] == '"':
            symbols.append(s[1:-1])
          elif s[0] == 'x' and s[1].isdigit():
            if len(s) > 3 and not hiero:
              symbols.append('[%s]' % s[3:])
            else:
              symbols.append('[X]')
        return symbols

    @staticmethod
    def getTerms(symbols):
      terms = []
      if type(symbols) == str:
          symbols = symbols.split(' ')
      for s in symbols:
        if len(s) < 2:
          if s == "@":
            break
        elif s[0] == '"' and s[-1] == '"':
          terms.append(s[1:-1])
      return terms

def getFlattenSymbols(symbols, grammar = 'scfg'):
    if isTree(symbols):
        # strSymbols is tree (S-expression)
        newSymbols = []
        tag = symbols[0]
        for s in symbols:
            if s[0:1] == '"':
                newSymbols.append(s)
            elif s[0:1] == "x":
                if grammar == 'scfg':
                    newSymbols.append(s)
                else:
                    fields = s.split(':')
                    newSymbols.append(fields[0]+':X')
        newSymbols.append("@")
        if grammar == 'hiero':
            newSymbols.append('X')
        else:
            newSymbols.append(tag)
        symbols = newSymbols
    return symbols

cpdef str getStrTravatarFeatures(dict dict_features):
    '''convert features dict into string in format of 'key=val' list'''
    cdef list featureList = []
    cdef str key
    cdef object val
    for key, val in dict_features.items():
        if key.find('egf') >= 0 or key.find('fge') >= 0:
            try:
                #val = math.log(val)
                val = round(math.log(val), 6)
            except:
                logging.warn( (key,val) )
        featureList.append( "%s=%s" % (key, val) )
    return str.join(' ', sorted(featureList))

def getTravatarFeatures(field):
  features = {}
  for strKeyVal in field.split():
    #debug.log(strKeyVal)
    (key, val) = strKeyVal.split('=')
    val = getNumber(val)
#    if key in ['egfl', 'egfp', 'fgel', 'fgep']:
#      val = math.e ** val
#    if key[-1] not in ['p', 'w']:
#      val = math.e ** val
    if len(key) >= 4 :
      val = math.e ** val
    features[key] = val
  return features

def isTree(symbols):
#    if len(symbols) >= 4 and symbols[0][0] != '"':
    if len(symbols) >= 4 and symbols[1] == '(':
        return True
    else:
        return False

def syncTags(src_symbols, trgSymbols, sync):
    if isTree(src_symbols):
        # src is tree (s-expression)
        if len(trgSymbols) >= 2 and trgSymbols[-2] == "@":
            # trg is tagged
            pass
        else:
            # trg is not tagged
            srcTag = src_symbols[0]
            srcNonTerminals = []
            for s in src_symbols:
                if s[0:1] == 'x' and s[1:2].isdigit():
                    srcNonTerminals.append(s)
#                        if len(s) > 3:
#                            srcNonTerminals.append(s[3:])
#                        else:
#                            srcNonTerminals.append('X')
            for i, s in enumerate(trgSymbols):
#                    print("TRG SYMBOLS: %s" % s)
                if s[0:1] == 'x' and s[1:2].isdigit():
#                        print("NONTERMINAL!")
                    if len(s) > 3:
                        # symbol is tagged
                        pass
                    else:
                        # symbol is not tagged
                        index = int(s[1])
                        if sync == 'scfg':
                            trgSymbols[i] = srcNonTerminals[index]
                        elif sync == 'hiero':
                            trgSymbols[i] = s + ":X"
#                self.trg = intern(self.trg + " @ " + srcTag)
            trgSymbols.append("@")
            if sync == 'scfg':
                trgSymbols.append(srcTag)
            elif sync == 'hiero':
                trgSymbols.append('X')
#            self.trg = intern(str.join(' ', trgSymbols))
#    if len(src_symbols) == 3:
#        tag = src_symbols[2]
#        if src_symbols[0] == "x0:"+tag:
#            # src is unary cycle
#            self.src = ""
#    print(trgSymbols)
    return src_symbols, trgSymbols




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Standard libraries
import argparse
import sys

# Local libraries
from lpu.common import logging
from lpu.common.files import wait_files

logger = logging.getColorLogger(__name__)

def cmd_wait_files(args):
    parser = argparse.ArgumentParser(description='wait until file will be found')
    parser.add_argument('filepaths', metavar='filepath', nargs='+', help='file path for waiting')
    parser.add_argument('--quiet', '-q', action='store_true', help='quiet mode')
    parser.add_argument('--debug', '-D', action='store_true', help='debug mode')
    parser.add_argument('--delay', '-d', default=0, type=float, help='duration to waiting time before the first check (default: %(default)s)')
    parser.add_argument('--interval', '-i', default=1, type=float, help='interval for next trial (default: %(default)s seconds)')
    parser.add_argument('--timeout', '-t', default=0, type=float, help='time limit in waiting file')
    parsed = parser.parse_args(args)

    with logging.using_config(logger) as c:
        if parsed.debug:
            c.set_debug(True)
            c.set_quiet(False)
        if parsed.quiet:
            c.set_quiet(True)
            c.set_debug(False)
        logger.debug(parsed)
        #waitFiles(**vars(parsed))
        wait_files(parsed.filepaths, parsed.interval, parsed.timeout, parsed.quiet)

def main():
    cmd_wait_files(sys.argv[1:])

if __name__ == '__main__':
    main()




# distutils: language=c++
# -*- coding: utf-8 -*-

'''this module provides globally shared stack of environment'''

# C++ setting
from libcpp cimport bool

# Standard libraries
import os

# Local libraries
import lpu
from lpu.__system__ import logging

logger = logging.getLogger(__name__)

env_stack = []

cdef _safe_debug_print(msg):
    try:
        #logger.debug(msg, stack_info=True)
        logger.debug(msg)
    except Exception as e:
        pass

def get_env(key, default=None, system=True):
    for env_dict in env_stack[::-1]:
        if key in env_dict:
            return env_dict[key]
    if system:
        if key in os.environ:
            return os.environ[key]
    else:
        return default

cdef class StackHolder:
    cdef readonly bool affect_system
    cdef readonly int level
    cdef dict env_layer
    cdef list back_log

    def __cinit__(self, affect_system=True):
        _safe_debug_print('initializing environ stack')
        self.env_layer = {}
        env_stack.append(self.env_layer)
        self.back_log = []
        self.affect_system = affect_system
        self.level = len(env_stack)

    cpdef set(self, str key, str value):
        cdef bool prev_exist = False
        cdef str prev_value = ''
        if self.affect_system:
            _safe_debug_print("setting %s='%s' in env" % (key, value) )
            if key in os.environ:
                prev_exist = True
                prev_value = os.environ[key]
            self.back_log.append( (key,prev_exist,prev_value) )
            os.environ[key] = value
        self.env_layer[key] = value

    cpdef clear(self):
        cdef str key
        cdef bool prev_exist
        cdef str prev_value
        if self.back_log:
            for key, prev_exist, prev_value in self.back_log[::-1]:
                if prev_exist:
                    _safe_debug_print("record back %s='%s' to env" % (key, prev_value))
                    if os and os.environ:
                        os.environ[key] = prev_value
                else:
                    _safe_debug_print("unset key from env: %s" % (key,))
                    try:
                        if os and os.environ:
                            os.environ.pop(key)
                    except Exception as e:
                        #logging.log(e)
                        #logger.debug(e)
                        pass
        self.env_layer.clear()
        # note: python2.7 does not have list.clear
        del self.back_log[:]

    def __enter__(self):
        logger.debug("entering environ stack")
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        _safe_debug_print("exiting from environ stack")
        #logging.debug(exception_type)
        #logging.debug(exception_value)
        #logging.debug(traceback)
        self.clear()

    def __dealloc__(self):
        _safe_debug_print("deallocating environ stack")
        self.clear()

def push(Class=StackHolder, **args):
    #env = StackHolder()
    env = Class()
    for key, val in args.items():
        env.set(key, str(val))
    return env




# distutils: language=c++
# -*- coding: utf-8 -*-

'''Customizable logging functions'''

# Standard libraries
import ast
import inspect
#import linecache
import re
import sys
import traceback

from lpu.__system__ import logging

#import lpu
from lpu.common import environ
from lpu.common import validation
from lpu.common.colors import put_color
from lpu.common.compat import MethodType

logger = logging.getLogger(__name__)

class LoggingStatus(environ.StackHolder):
    def __init__(self, logger=None):
        super(LoggingStatus,self).__init__()
        self.loggers = None

    def _reconfigureLogger(self):
        if self.loggers:
            for logger in self.loggers:
                configureLogger(logger)
        else:
            try:
                import lpu
                configureLogger(lpu.logger)
            except Exception as e:
                pass

    def set_loggers(self, loggers):
        if loggers:
            set_loggers = set()
            if not isinstance(loggers, (list,tuple)):
                loggers = [loggers]
            for logger in loggers:
                if isinstance(logger, str):
                    logger = logging.getLogger(logger)
                set_loggers.add(logger)
            self.loggers = set_loggers

    def set_debug(self, enable=True):
        if enable:
            self.set('DEBUG', '1')
        else:
            self.set('DEBUG', '0')
        self._reconfigureLogger()
    def unset_debug(self):
        return self.clear('DEBUG')

    def set_quiet(self, enable=True):
        if enable:
            self.set('QUIET', '1')
        else:
            self.set('QUIET', '0')
        self._reconfigureLogger()
    def unset_debug(self):
        return self.clear('QUIET')

    def __enter__(self):
        logger.debug("entering logging environment")
        super(LoggingStatus,self).__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        super(LoggingStatus,self).__exit__(exc_type, exc_val, exc_tb)
        logger.debug("exiting from logging environment")
        self._reconfigureLogger()

def get_debug_status():
    mode = environ.get_env('DEBUG')
    if not mode:
        return False
    elif mode.lower() in ('', 'false', 'off', '0'):
        return False
    else:
        return True

def get_color_status():
    mode = environ.get_env('COLOR')
    auto = False
    if not mode:
        auto = True
    else:
        if mode.lower() in ('false', 'off', '0'):
            return False
        elif mode.lower() in ('true', 'on', 'always', 'force', '1'):
            return True
        elif mode.lower() in ('', 'auto',):
            auto = True
    if auto:
        return sys.stderr.isatty()
    else:
        return False

def get_quiet_status():
    mode = environ.get_env('QUIET')
    if not mode:
        return False
    else:
        if mode.lower() in ('', 'false', '0'):
            return False
        elif mode.lower() in ('true', '1'):
            return True
    return False

class FilterCondition(logging.Filter):
    def __init__(self, **rules):
        super(FilterCondition,self).__init__()
        self.rules = rules

    def filter(self, record):
        if 'level' in self.rules:
            if self.rules['level'] not in [record.levelname, record.levelno]:
                return False
        return True

def getLevelString(level):
    if isinstance(level, int):
        return logging.getLevelName(level)
    elif isinstance(level, str):
        return level
    else:
        validation.check_argument_type(level, 'level', (int, str))

class ColorizingFormatter(logging.Formatter):
    def __init__(self, fmt=None, datefmt=None):
        super(ColorizingFormatter,self).__init__(fmt, datefmt)
        self._format_rules = []
        #self._default_fmt = fmt
        self._default_fmt = self._fmt
        self._colors = dict()

    def addFormatRule(self, rule, fmt=None):
        #print("adding format: {}".format(fmt))self._fmt
        #print("adding filter: {}".format(filter))
        #self._format_rules.append([rule, fmt])
        self._format_rules.insert(0, [rule, fmt])

    def _setColorizedFormat(self, fmt, record):
        self._fmt = fmt
        if sys.version_info.major == 3:
            self._style._fmt = fmt
        return fmt

    #def _colorizeText(self, record, text):
    #cpdef str _colorizeText(self, record, str text):
    def _colorizeText(self, record, str text):
        cdef str color_default
        level = record.levelname.lower()
        color_level = self._colors.get(level, None)
        if color_level:
            text = put_color(text, color_level)
        else:
            color_default = self._colors.get('default', None)
            if color_default:
                text = put_color(text, color_default)
        return text

    def format(self, record):
        #print("formatting... {}".format(record))
        fmt_apply = None
        for flt, fmt in self._format_rules:
            if flt.filter(record):
                #print("filter: {}".format(flt))
                #print("format: {}".format(fmt))
                #print("record: {}".format(record))
                if fmt:
                    #self._style._fmt = self._fmt = fmt
                    #setFormat(self, fmt)
                    #self._setColorizedFormat(fmt, record)
                    fmt_apply = fmt
                break
        if fmt_apply:
            self._setColorizedFormat(fmt_apply, record)
        else:
            self._setColorizedFormat(self._default_fmt, record)
        text = super(ColorizingFormatter,self).format(record)
        return self._colorizeText(record, text)

    def formatStack(self, stack_info):
        #logger.debug(stack_info)
        formatted = stack_info.rstrip()
        formatted = '  ' + formatted.replace('\n', '\n  ')
        color_stack = self._colors.get('stack', None)
        if color_stack:
            formatted = put_color(formatted, color_stack)
        else:
            color_debug = self._colors.get('debug', None)
            if color_debug:
                formatted = put_color(formatted, color_debug)
        return formatted

    def formatException(self, exc_info):
        #logger.debug(exc_info)
        etype, value, tb = exc_info
        list_formatted = traceback.format_exception(etype, value, tb)
        formatted = str.join('', list_formatted).rstrip()
        formatted = '  ' + formatted.replace('\n', '\n  ')
        #formatted = '  ' + str.join('  ', list_formatted).rstrip()
        color_exception = self._colors.get('exception', None)
        if color_exception:
            formatted = put_color(formatted, color_exception)
        else:
            color_error = self._colors.get('error', None)
            if color_error:
                formatted = put_color(formatted, color_error)
        return formatted

    def setColor(self, keyword, color_name):
        #if isinstance(keyword, str):
        if isinstance(keyword, (str,bytes)):
            keyword = keyword.lower()
            keyword = keyword.replace('color_', '')
        else:
            #validation.check_argument_type(keyword, 'keyword', str)
            validation.check_argument_type(keyword, 'keyword', (str,bytes))
        if color_name is None:
            # unsetting
            if keyword in self._colors:
                del self._colors[keyword]
        elif isinstance(color_name, str):
            color_name = color_name.lower()
            self._colors[keyword] = color_name
        else:
            validation.check_argument_type(color_name, 'color_name', str)

    def setColors(self, **kwargs):
        for key, color in kwargs.items():
            self.setColor(key, color)

    def setLevelFormat(self, level, fmt):
        level_name = getLevelString(level)
        level_rule = FilterCondition(level = level)
        self.addFormatRule(level_rule, fmt)

    def setLevelColor(self, level, color_name):
        level_name = getLevelString(level)
        #self._colors[level_name] = color_name
        self.setColor(level_name, color_name)

DEFAULT_DATE_FORMAT = "%Y/%m/%d %H:%M:%S"
DEFAULT_FORMAT = "[%(asctime)s %(name)s] %(message)s"
DEFAULT_DEBUG_FORMAT = "[%(asctime)s %(name)s %(filename)s:%(funcName)s:%(lineno)s] %(message)s"
DEFAULT_INFO_FORMAT     = DEFAULT_FORMAT
DEFAULT_WARNING_FORMAT  = DEFAULT_FORMAT
DEFAULT_ERROR_FORMAT    = DEFAULT_FORMAT
DEFAULT_CRITICAL_FORMAT = DEFAULT_FORMAT
DEFAULT_DEBUG_COLOR = 'yellow'
DEFAULT_INFO_COLOR  = 'cyan'
#DEFAULT_WARNING_COLOR  = 'purple'
DEFAULT_WARNING_COLOR  = 'yellow'
DEFAULT_ERROR_COLOR    = 'yellow'
DEFAULT_CRITICAL_COLOR = 'red'

def colorizeHandler(handler, mode='auto'):
    # deciding whether to enable colorizing mode (based on arguments, environment)
    if mode == 'auto':
        if isinstance(handler.formatter, ColorizingFormatter):
            # already colorized (itself or any ancestor)
            return handler
        enable = get_color_status()
    else:
        enable = str(mode).lower() in ('1', 'on', 'true', 'always', 'force')
    # setting colorizing formatter
    formatter = ColorizingFormatter(DEFAULT_FORMAT, DEFAULT_DATE_FORMAT)
    formatter.setLevelFormat(logging.DEBUG,    DEFAULT_DEBUG_FORMAT)
    formatter.setLevelFormat(logging.INFO,     DEFAULT_INFO_FORMAT)
    formatter.setLevelFormat(logging.WARNING,  DEFAULT_WARNING_FORMAT)
    formatter.setLevelFormat(logging.ERROR,    DEFAULT_ERROR_FORMAT)
    formatter.setLevelFormat(logging.CRITICAL, DEFAULT_CRITICAL_FORMAT)
    if enable:
        formatter.setLevelColor(logging.DEBUG,    DEFAULT_DEBUG_COLOR)
        formatter.setLevelColor(logging.INFO,     DEFAULT_INFO_COLOR)
        formatter.setLevelColor(logging.WARNING,  DEFAULT_WARNING_COLOR)
        formatter.setLevelColor(logging.ERROR,    DEFAULT_ERROR_COLOR)
        formatter.setLevelColor(logging.CRITICAL, DEFAULT_CRITICAL_COLOR)
    handler.setFormatter(formatter)
    return formatter

cdef _checkLoggerColorized(logger):
    while logger:
        for handler in logger.handlers:
            if isinstance(handler.formatter, ColorizingFormatter):
                return True
        # checking ancestors
        logger = logger.parent
    return False

def colorizeLogger(logger, mode='auto'):
    if isinstance(logger, str):
        logger = getLogger(logger)
    handlers = logger.handlers
    for handler in handlers:
        colorizeHandler(handler, mode)
    return logger

def colorize(obj):
    if isinstance(obj, logging.Logger):
        return colorizeLogger(obj)
    elif isinstance(obj, logging.Handler):
        return colorizeHandler(obj)

def configureLogger(logger, mode='auto'):
    if not logger:
        return logger
    if mode == 'auto':
        if get_quiet_status():
            logger.setLevel(logging.ERROR)
        elif get_debug_status():
            logger.setLevel(logging.DEBUG)
        else:
            # verbose mode
            logger.setLevel(logging.INFO)
    else:
        logger.setLevel(mode)
    return logger

_cache = {}
cdef _get_cached_line(path, lineno, fallback):
    try:
        if path not in _cache:
            _cache[path] = open(path).readlines()
        if lineno in range(1, len(_cache[path])+1):
            return _cache[path][lineno-1]
    except:
        pass
    return fallback

cpdef _debug_print(self, val=None, limit=0):
    stack = traceback.extract_stack(limit=limit)
    format = ""
    if limit > 0:
        format = ""
        for path, lineno, func, line in stack:
            line = _get_cached_line(path, lineno, line).strip()
            s = '\n  file:{}, line:{}, func:{}, code:{}'.format(path, lineno, func, line)
            format += s
    stack = traceback.extract_stack(limit=1)
    path, lineno, func, line = stack[0]
    line = _get_cached_line(path, lineno, line).strip()
    if val is not None:
        expr = re.findall(r'\(.*\)$', line)
        if expr:
            expr = expr[0][1:-1].strip()
            if expr.find(',') > 0:
                expr = str.join(',', expr.split(',')[:-1]).strip()
            if isinstance(val, (int,float)):
                str_val = '{}({})'.format(type(val).__name__,val)
            elif isinstance(val, str):
                str_val = val
            else:
                str_val = repr(val)
            if str_val.find('\n') >= 0:
                # multiple lines
                format = "{} => (see following lines)\n{}".format(expr, str_val) + format
            else:
                # single line
                format = "{} => {}".format(expr, str_val) + format
        else:
            format = str(val) + format
    #module_name = inspect.getmodulename(path)
    #if not module_name:
    #    module_name = '__main__'
    #module_name = '__main__'
    #logger = getColorLogger(module_name)
    #logger = getLogger(module_name)
    #logger = colorizeLogger(logger)
    #logger.debug(format)
    self.debug(format)

def debug_print(val=None, limit=0):
    logger = getColorLogger('__main__')
    return logger.debug_print(val, limit)

def getColorLogger(name, level_mode='auto', add_handler='auto'):
    logger = logging.getLogger(name)
    if level_mode is not 'auto':
        logger = configureLogger(logger, mode=level_mode)
    logger.debug_print = MethodType(_debug_print, logger, logging.Logger)
    if add_handler == 'auto':
        if _checkLoggerColorized(logger):
            add_handler = None
        else:
            add_handler = logging.StreamHandler()
    if add_handler:
        # duplication check
        for handler in logger.handlers:
            if handler is add_handler:
                add_handler = None
        if add_handler:
            logger.addHandler(add_handler)
            if level_mode is 'auto':
                configureLogger(logger, mode=level_mode)
    return colorizeLogger(logger)

# global environ
#def push_environ(logger):
#cpdef using_config(logger, debug=None, quiet=None):
cpdef using_config(loggers, debug=None, quiet=None):
    env_layer = environ.push(LoggingStatus)
    env_layer.set_loggers(loggers)
    if debug is not None:
        env_layer.set_debug(debug)
    if quiet is not None:
        env_layer.set_quiet(debug)
    return env_layer
#env = push_environ(logger)
env = using_config(None)

# importing from system logging module
NOTSET   = logging.NOTSET
INFO     = logging.INFO
DEBUG    = logging.DEBUG
WARNING  = logging.WARNING
ERROR    = logging.ERROR
CRITICAL = logging.CRITICAL

Filter = logging.Filter
StreamHandler = logging.StreamHandler
FileHandler = logging.FileHandler

getLogger = logging.getLogger




# distutils: language=c++
# -*- coding: utf-8 -*-

'''classes handling phrase/rule table'''

from libcpp cimport bool

# 3rd party library
import pycedar

# Local libraries
from lpu.common import logging
from lpu.common import progress
from lpu.smt.trans_models import records

key_types = ['src', 'src_hiero', 'src_symbols', 'src_tree']

#cdef get_track_str(object trie, str key):
#    return str(trie.get_node(key)).replace(' ', '')
    #return str(node.track()).replace(' ', '')
cdef get_track_str(object node):
    return str(node.track()).replace(' ', '')

cdef track2key(object trie_dict, str track):
    cdef list numbers
    cdef size_t node_id, length
    numbers = track[1:-1].split(',')
    node_id = int(numbers[0])
    length  = int(numbers[1])
    return trie_dict.trie.suffix(node_id, length)

cdef class Table(object):
    cdef readonly object RecordClass
    cdef str table_path
    cdef object table_file
    cdef str key_type
    cdef readonly object field_dict
    #cdef object record_dict
    cdef readonly object record_dict
    #cdef readonly object trg_record_dict
    #cdef readonly object key_record_dict
    cdef readonly object src_record_dict
    cdef readonly object trg_record_dict
    cdef readonly object trg_field_record_dicts
    cdef bool trg_key_enabled

    def __init__(self, table_path, RecordClass, trg_key = False, **options):
        self.key_type = options.get('key_type', 'src')
        self.RecordClass = RecordClass
        self.table_path = table_path
        self.field_dict  = pycedar.dict(str)
        self.record_dict = pycedar.dict(str)
        #self.key_record_dict    = pycedar.dict(str)
        self.src_record_dict    = pycedar.dict(str)
        #if trg_key:
        #    self.trg_record_dict = pycedar.dict(str)
        #else:
        #    self.trg_record_dict = None
        self.trg_key_enabled = trg_key
        if trg_key:
            self.trg_record_dict = pycedar.dict(str)
            self.trg_field_record_dicts = []
        self.__load()

    #cdef add_key_and_node(self, str key, object record_node):
    #    cdef str key_track, pair_tracks
    #    key_track = get_track_str(self.field_dict.get_node(key))
    #    pair_tracks = "%s | %s" % (key_track, get_track_str(record_node))
    #    self.key_record_dict.update(pair_tracks, 1)

    cdef add_key_and_node(self, object key_record_dict, str key, object record_node):
        cdef str key_track, pair_tracks
        key_track = get_track_str(self.field_dict.get_node(key))
        pair_tracks = "%s | %s" % (key_track, get_track_str(record_node))
        #self.key_record_dict.update(pair_tracks, 1)
        key_record_dict.update(pair_tracks, 1)

    cdef add_fields(self, str line):
        cdef str field
        for field in line.split('|||'):
            field = field.strip()
            if field:
                try:
                    self.field_dict.update(field, 1)
                except Exception as e:
                    logging.debug(e)
                    logging.debug(field)

    cpdef str format_src_key(self, str src):
        if self.key_type == 'src':
            #return src + ' ||| '
            return src
        elif self.key_type == 'src_hiero':
            #return str.join(' ', self.RecordClass.getSymbols(src,hiero=True)) + ' ||| '
            return str.join(' ', self.RecordClass.getSymbols(src,hiero=True))
        elif self.key_type == 'src_symbols':
            #return str.join(' ', self.RecordClass.getSymbols(src,hiero=False)) + ' ||| '
            return str.join(' ', self.RecordClass.getSymbols(src,hiero=False))

    cpdef str line2tracks(self, str line):
        cdef list track_list = []
        cdef str str_track
        for field in line.split('|||'):
            field = field.strip()
            if len(field) == 0:
                track_list.append('')
            elif field in self.field_dict:
                #str_track = get_track_str(self.field_dict, field)
                str_track = get_track_str( self.field_dict.get_node(field) )
                #str_track = str( self.field_dict.get_node(field).track() ).replace(' ', '')
                track_list.append( str_track )
            else:
                #logging.warn("'%s' is not in self.field_dict" % field)
                track_list.append( "(-1,-1)" )
        return str.join(' | ', track_list)

    cpdef str tracks2line(self, str tracks):
    #cpdef str tracks2line(self, str tracks):
        cdef list field_list = []
        #cdef (int,int) numbers
        cdef list numbers
        cdef size_t node_id, length
        for track in tracks.split('|'):
            track = track.strip()
            if track:
                #field is form of "(id,length)")
                #numbers = ( field[1:-1].split(',') )
                #node_id = int(numbers[0])
                #length  = int(numbers[1])
                #field_list.append( self.field_dict.trie.suffix(node_id,length) )
                field_list.append( track2key(self.field_dict, track) )
            else:
                field_list.append( '' )
        return str.join(' ||| ', field_list)

    cdef __load(self):
        cdef long i
        cdef str line
        #cdef object rec
        cdef str src, trg
        cdef str srcKey
        #print(self.table_file)
        cdef str line_tracks
        cdef str src_tracks
        cdef str trg_tracks
        cdef str str_track_pair
        cdef str pair_tracks
        #cdef list rec_tracks
        cdef object node
        cdef str field

        for i, line in enumerate(progress.view(self.table_path, 'building trie of all fields')):
            #if i > 200000:
            #    break
            try:
                src = line.split('|||',1)[0].strip()
                self.add_fields(line)
                self.add_fields(self.format_src_key(src))
                if self.trg_key_enabled:
                    trg = line.split('|||',2)[1].strip()
                    self.add_fields(trg)
                    if trg.find('|COL|') >= 0:
                        for field in trg.split('|COL|'):
                            self.add_fields(field.strip())
            except Exception as e:
                logging.warn("file: %s, line: %s" % (self.table_path, i+1))
                logging.warn(e)
                raise e

        for i, line in enumerate(progress.view(self.table_path, 'building trie of all records')):
            try:
                line_tracks = self.line2tracks(line.strip())
                self.record_dict.update(line_tracks, 1)
            except Exception as e:
                logging.warn("file: %s, line: %s" % (self.table_path, i+1))
                logging.warn(e)
                raise e

        for node in progress.view(self.record_dict.nodes(), 'registering pair of (key,record)', max_count=len(self.record_dict)):
            try:
                line_tracks = node.key()
                line = self.tracks2line(line_tracks)
                src = line.split('|||',1)[0].strip()
                src_key = self.format_src_key(src)
                #self.add_key_and_node(src_key, node)
                self.add_key_and_node(self.src_record_dict, src_key, node)
                if self.trg_key_enabled:
                    trg = line.split('|||', 2)[1].strip()
                    #self.add_key_and_node(trg, node)
                    self.add_key_and_node(self.trg_record_dict, trg, node)
                    if trg.find('|COL|') >= 0:
                        for i, field in enumerate(trg.split('|COL|')):
                            if i >= len(self.trg_field_record_dicts):
                                self.trg_field_record_dicts.append( pycedar.dict(str) )
                            #self.add_key_and_node(field.strip(), node)
                            self.add_key_and_node(self.trg_field_record_dicts[i], field.strip(), node)
            except Exception as e:
                logging.warn("line_tracks: %s" % line_tracks)
                logging.warn("line: %s" % line)
                #logging.warn("src_key: %s" % src_key)
                logging.warn(e)
                raise e

    def __find_key(self, object key_record_dict, str key, bool force=True):
        cdef str line
        cdef str key_track
        cdef str pair_tracks
        cdef str record_track
        cdef str line_tracks

        key_track = self.line2tracks(key)
        if key_track.find('(-1,-1)') >= 0:
            return
        for pair_tracks in key_record_dict.find_keys(key_track,force=force):
            record_track = pair_tracks.split('|',1)[1].strip()
            line_tracks = track2key(self.record_dict, record_track)
            line = self.tracks2line(line_tracks)
            yield self.RecordClass(line)

    def find(self, str prefix, bool force=True):
        cdef str prefix_tracks
        cdef str line_tracks
        cdef str line

        prefix_tracks = self.line2tracks(prefix)
        for line_tracks in self.record_dict.find_keys(prefix_tracks, force=force):
            line = self.tracks2line(line_tracks)
            yield self.RecordClass(line)

    #def find(self, str key):
    #    cdef str line
    #    #cdef str str_tracks
    #    cdef str key_track
    #    cdef str pair_tracks
    #    cdef str record_track
    #    cdef str line_tracks

    #    if key:
    #        #str_tracks = self.line2tracks(key)
    #        key_track = self.line2tracks(key)
    #        if key_track.find('(-1,-1)') >= 0:
    #            return
    #        #for line in self.record_dict.find_keys(str_tracks, force=True):
    #        #    main_tracks = line.split('|||', 1)[1]
    #        #    yield self.RecordClass( self.tracks2line(main_tracks) )
    #        #for pair_tracks in self.key_record_dict.find_keys(key_track, force=True):
    #        for pair_tracks in self.src_record_dict.find_keys(key_track, force=True):
    #            record_track = pair_tracks.split('|',1)[1].strip()
    #            line_tracks  = track2key(self.record_dict, record_track)
    #            line = self.tracks2line(line_tracks)
    #            yield self.RecordClass( line )
    #    else:
    #        # empty key, find all records (without key)
    #        for line_tracks in self.record_dict.find_keys('', force=True):
    #            line = self.tracks2line(line_tracks)
    #            yield self.RecordClass( line )

    #cpdef find_src(self, str src):
        #return self.find(self.format_src_key(src))

    cpdef find_src(self, str src, force=True):
        return self.__find_key(self.src_record_dict, self.format_src_key(src), force=force)

    #cpdef find_trg(self, str trg):
    #    if not trg:
    #        raise KeyError("trg should not be empty")
    #    return self.find(trg)

    cpdef find_trg(self, str trg, object target_index=None, force=True):
        #if target_number >= 0:
        if isinstance(target_index,int):
            return self.__find_key(self.trg_field_record_dicts[target_index], trg, force=force)
        else:
            return self.__find_key(self.trg_record_dict, trg, force=force)

    #def find_trg(self, str trg):
    #    if self.trg_record_dict is None:
    #        raise LookupError('target key is not enabled, please init with Table(..., trg_key=True)')
    #    cdef str line
    #    cdef str str_tracks
    #    str_tracks = self.line2tracks(trg)

    #    if str_tracks == '(-1,-1)':
    #        return
    #    for line in self.trg_record_dict.find_keys(str_tracks, force=True):
    #        main_tracks = line.split('|||', 1)[1]
    #        yield self.RecordClass( self.tracks2line(main_tracks) )

    def __iter__(self):
        return self.find('')

    def __len__(self):
        return len(self.record_dict)

class MosesTable(Table):
    def __init__(self, table_path, **options):
        Table.__init__(self, table_path, records.MosesRecord, **options)

class TravatarTable(Table):
    def __init__(self, table_path, **options):
        Table.__init__(self, table_path, records.TravatarRecord, **options)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Standard libraries
import argparse
import math
import multiprocessing
import os
import subprocess
import sys
import time

# Local libraries
from lpu.common import files
from lpu.common import progress
from lpu.common import logging
from lpu.common.config import Config
from lpu.commands.wait_files import wait_file

logger = logging.getColorLogger(__name__)

numCPUs = multiprocessing.cpu_count()
SLEEP_DURATION = 1.0

#def getHostProcID():
def getCurrentWorkerID():
    return "%s:%s" % (os.uname()[1],os.getpid())

def report(filepath, message):
    if os.path.exists(filepath):
        logger.info('Exists file or directory: %s' % filepath)
        return False
    else:
        logger.info('Reporting into file: %s' % filepath)
        fobj = open(filepath, 'w')
        fobj.write(message)
        fobj.close()
        return True

def remove(f):
    if type(f) == str:
        logger.info('Removing file: %s' % f)
        os.remove(f)
    elif isinstance(f, files.FileType):
        logger.info('Removing file: %s' % f.name)
        f.close()
        os.remove(f.name)

def checkFile(filepath):
    if os.path.exists(filepath):
        logger.info('File already exists: %s' % filepath)
        return True
    else:
        return False

#def checkStage(tmpdir, basename, stage):
#def checkPhase(tmpdir, phase):
def checkPhase(conf, phase):
    tmpdir = conf.data.tmpdir
#    if checkFile('%s/__INIT__.%s.%s' % (tmpdir,stage,basename)):
    if checkFile('%s/report.%s.begin' % (tmpdir,phase)):
        return 'started'
#    elif checkFile('%s/__DONE__.%s.%s' % (tmpdir,stage,basename)):
    elif checkFile('%s/report.%s.done' % (tmpdir,phase)):
        return 'finished'
    else:
        return 'none'

def getPhaseCharge(conf, phase):
    tmpdir = conf.data.tmpdir
    #basename = conf.data.basename
    #path = '%(tmpdir)s/__INIT__.%(stage)s.%(basename)s' % locals()
    path = '%(tmpdir)s/report.%(phase)s.begin' % locals()
    #return open(path, 'r').read()
    chargeID = open(path, 'r').read()
    return chargeID

def checkPhaseCharge(conf, phase):
    if getPhaseCharge(conf,phase) != getCurrentWorkerID():
        logger.warning('Failed to confirm the responsible process of the phase: "%s"' % phase)
        return False
    return True

def reportInit(conf, phase):
    tmpdir = conf.data.tmpdir
    #basename = conf.data.basename
    #path = '%(tmpdir)s/__INIT__.%(stage)s.%(basename)s' % locals()
    path = '%(tmpdir)s/report.%(phase)s.begin' % locals()
    #if not report(path, conf.data.hostproc):
    if not report(path, getCurrentWorkerID()):
        return False
    logger.info('Waiting %s second to confirm the responsible process of the phase' % conf.data.interval)
    time.sleep(conf.data.interval)
    return checkPhaseCharge(conf, phase)
    #return True

def reportDone(conf, phase):
    tmpdir = conf.data.tmpdir
    #basename = conf.data.basename
    #return report('%(tmpdir)s/__DONE__.%(stage)s.%(basename)s'%locals(), conf.data.hostproc)
    return report('%(tmpdir)s/report.%(phase)s.done'%locals(), getCurrentWorkerID())

def waitPhaseDone(conf, phase):
    tmpdir = conf.data.tmpdir
    #basename = conf.data.basename
    #return waitFile('%(tmpdir)s/__DONE__.%(stage)s.%(basename)s'%locals())
    return wait_file('%(tmpdir)s/report.%(phase)s.done' % locals())

def getInBuffer(conf):
    #bufname = '%s/__BUFFER__%s' % (tmpdir,hostproc)
    #bufname = '%(tmpdir)s/__BUFFER__%(hostproc)s' % conf
    bufname = '%(tmpdir)s/tmp.buffer' % conf
    #progCounter = progress.ProgressCounter(1, "buffering", force=True)
    with open(conf.data.inPath, 'r') as inFile:
        inbuf = open(bufname, 'w+')
        logger.info("Buffering into file: \"%s\"" % inbuf.name)
        lineCount = 0
        #for line in inFile:
        with progress.view(inFile, 'buffering') as p:
            for line in p:
                lineCount += 1
                inbuf.write(line)
    conf.data.lineCount = lineCount
    #progCounter.flush()
    logger.info("Lines: %s" % lineCount)
    inbuf.seek(0)
    return inbuf

def int2str(number, digits, suppress='0'):
    strNumber = str(number)
    lenNumber = len(strNumber)
    return suppress*(digits-lenNumber) + strNumber

def getSplitPrefix(conf):
    #return "%(tmpdir)s/%(basename)s" % conf
    return "%(tmpdir)s/split" % conf

def splitFile(conf):
    configFile = "%s/config.json" % conf.data.tmpdir
    if not reportInit(conf, 'split'):
        waitPhaseDone(conf, 'split')
        logger.debug(conf)
        logger.info("Updating the configuration with: %s" % configFile)
        conf.load_json(open(configFile).read(), False)
        logger.debug(conf)
        return True
    #threads = conf.require('threads')
    splitSize = conf.get('splitSize', None)
    #numChunks = conf.data.numChunks
    numChunks = conf.get('numChunks', conf.data.threads)
    inbuf = getInBuffer(conf)
    lineCount = conf.data.lineCount
    if lineCount == 0:
        logger.info('Nothing to do')
        remove(inbuf)
        return
    #prefix = getPrefix(conf)
    #strSplitPrefix = conf.data.tmpdir + "/split"
    prefix = getSplitPrefix(conf)
    if not splitSize:
        #splitSize = int( math.ceil(float(lineCount) / threads) )
        splitSize = conf.data.splitSize = int( math.ceil(float(lineCount) / numChunks) )
        logger.info('Split size: %s' % splitSize)
    #splitCount = conf.data.splitCount = int(math.ceil(float(lineCount) / splitSize))
    #splitCount = int(math.ceil(float(lineCount) / splitSize))
    numChunks = conf.data.numChunks = int(math.ceil(float(lineCount) / splitSize))
    #digits = conf.data.digits = len(str(splitCount))
    digits = conf.data.digits = len(str(numChunks))
    logger.info('Splitting into: "%s.*"' % prefix)
    #progCounter = progress.ProgressCounter(1, "splitting", force=True, maxCount=lineCount)
    progInbuf = progress.FileReader(inbuf, "splitting")
    with inbuf:
        for fileNumber in range(1, numChunks+1):
            path = "%s.%s.in" % (prefix, int2str(fileNumber,digits,'0'))
            with open(path,'w') as outFile:
                for _ in range(0, splitSize):
                    line = progInbuf.readline()
                    #line = inbuf.readline()
                    if line:
                        outFile.write(line)
                        #progCounter.add(1, view=True)
                    else:
                        break
    progInbuf.close()
    #progCounter.flush()
    logger.info('Finished to split')
    remove(inbuf)
    logger.info('Saving configuration into: %s' % configFile)
    with open(configFile, 'w') as fobj:
        fobj.write(conf.to_json(indent=4))
    reportDone(conf, 'split')
    return True

#def getFileNumberDigits(conf):
#    MAX_DIGITS=20
#    #prefix = getPrefix(conf)
#    prefix = getSplitPrefix(conf)
#    for digits in range(1, MAX_DIGITS+1):
#        path = "%s.%s.in" % (prefix,int2str(1, digits, '0'))
#        if os.path.exists(path):
#            return digits
#    logging.alert("Failed to get file number digits")

def waitAvailableWorker(conf, workers, flush = False):
    threads = min(conf.data.threads, numCPUs)
    if flush:
        threads = 1
    if len(workers) == 0:
        return
    while True:
        for i, (proc,phase) in enumerate(workers):
            if proc.poll() != None:
                conf.data.processed += 1
                reportDone(conf,phase)
                workers.pop(i)
                break
        else:
            time.sleep(conf.data.interval)
        if len(workers) < threads:
            break
    processed = conf.data.processed
    numChunks = conf.data.numChunks
    ratio = float(processed) / numChunks
    strTemplate = "Processed files: %s / %s (%2.2f%%), Active processes: %s / %s"
    strMessage = strTemplate % (processed,numChunks,ratio*100,len(workers),threads)
    logger.info(strMessage)

def runWorkers(conf):
    #fileNumber = 1
    #prefix = getPrefix(conf)
    prefix = getSplitPrefix(conf)
    #digits = getFileNumberDigits(conf)
    numChunks = conf.data.numChunks
    digits = conf.data.digits = len(str(numChunks))
    #threads = min(conf.data.threads,numCPUs)
    workers = []
    conf.data.processed = 0
    for fileNumber in range(1, numChunks+1):
        strFileNumber = int2str(fileNumber, digits, '0')
        strPhase = 'cmd.%s' % strFileNumber
        inPath = "%s.%s.in" % (prefix, strFileNumber)
        outPath = "%s.%s.out" % (prefix, strFileNumber)
        #cmdline = "%s < %s > %s" % (conf.data.command, inPath, outPath)
        cmdline = 'cat "%s" | %s > "%s"' % (inPath, conf.data.command, outPath)
        waitAvailableWorker(conf, workers)
        if not reportInit(conf, strPhase):
            logger.info("Skipping processing: %s" % inPath)
            #if checkPhase(conf.data.tmpdir, strPhase) == 'finished':
            if checkPhase(conf, strPhase) == 'finished':
                conf.data.processed += 1
            continue
        proc = subprocess.Popen(cmdline, shell=True)
        workers.append([proc,strPhase])
        logger.info("Executing: %s" % cmdline)
        #logging.debug(p)
        #reportInit(conf, 'cmd.%s' % strFileNumber)
        fileNumber += 1
    waitAvailableWorker(conf, workers, flush=True)

def concatFiles(conf):
    if not reportInit(conf, 'concat'):
        waitPhaseDone(conf, 'concat')
        chargeID = getPhaseCharge(conf, 'concat')
        logger.info("Finalizing (concatenation) process is running: %s" % chargeID)
        return True
    prefix = getSplitPrefix(conf)
    numChunks = conf.data.numChunks
    lineCount = conf.data.lineCount
    digits = conf.data.digits = len(str(numChunks))
    #outPath = "%s.out" % (prefix)
    outPath = conf.data.outPath
    conf.data.processed = 0
    #progCounter = progress.ProgressCounter(1, "concat", force=True, maxCount=lineCount)
    progCounter = progress.SpeedCounter(header="concat", max_count=lineCount)
    for fileNumber in range(1, numChunks+1):
        strFileNumber = int2str(fileNumber, digits, '0')
        inPath = "%s.%s.out" % (prefix, strFileNumber)
        wait_file(inPath)
        conf.data.processed += 1
    logger.info('Concatenating: "%s.*" -> "%s"' % (prefix,outPath))
    with open(outPath, 'w') as outFile:
        for fileNumber in range(1, numChunks+1):
            strFileNumber = int2str(fileNumber, digits, '0')
            inPath = "%s.%s.out" % (prefix, strFileNumber)
            with open(inPath, 'r') as inFile:
                for line in inFile:
                    outFile.write(line)
                    progCounter.add(1, view=True)
    progCounter.reset()
    reportDone(conf, 'concat')

def checkConfig(conf):
    conf.setdefault('inPath',  '/dev/stdin')
    conf.data.inPath = os.path.abspath(conf.data.inPath)
    conf.setdefault('outPath', '/dev/stdout')
    conf.data.outPath = os.path.abspath(conf.data.outPath)
    #conf.setdefault('splitSize', None)
    conf.setdefault('interval', SLEEP_DURATION)
    conf.setdefault('threads', numCPUs)
    #conf.setdefault('numChunks', conf.data.threads)
    if conf.get('numChunks', None):
        if conf.data.numChunks <= 0:
            strTemplate = "--chunks (number of splitted files) should be positive integer: %s"
            strMessage = strTemplate % (conf.data.numChunks)
            logger.error(strMessage)
    #conf.setdefault('basename', os.path.basename(conf.data.inPath))
    conf.setdefault('basename', os.path.basename(conf.data.outPath))
    conf.setdefault('tmpdir', './tmp-%s' % conf.data.basename)
    conf.data.tmpdir = os.path.abspath(conf.data.tmpdir)
    #if conf.data.threads > numCPUs:
    #    strTemplate = "Number of worker processes is limited to number of available threads: %s -> %s"
    #    strMessage = strTemplate % (conf.data.threads, numCPUs)
    #    logging.warn(strMessage)
    #    conf.data.threads = numCPUs

def execParallel(conf = None, **others):
    conf = Config(conf, **others)
    checkConfig(conf)
    logger.debug(conf)
    #hostproc = conf.data.hostproc = getHostProcID()
    workerID = getCurrentWorkerID()
    files.safeMakeDirs(conf.data.tmpdir)
    logger.info("Worker ID (Host+Proc): \"%s\"" % workerID)
    splitFile(conf)
    runWorkers(conf)
    concatFiles(conf)

def cmdExecParallel(args):
    parser = argparse.ArgumentParser(description='Execute command in multiple processes by splitting the target file')
    #parser.add_argument('inFile', type=str, help='input file name for execution')
    #parser.add_argument('outFile', type=str, help='output file name for execution')
    parser.add_argument('command', type=str, help='command line string for execution, replacing %%1 and %%2 with input and output files respectively')
    parser.add_argument('--input',  '-I', dest='inPath', type=str, default='/dev/stdin',  help='path to input file for execution (default: /dev/stdin)')
    parser.add_argument('--output', '-O', dest='outPath', type=str, default='/dev/stdout', help='path to output file for execution (default: /dev/stdout)')
    #parser.add_argument('--digitsize', '-d', type=int, default=6, help='assign the number of digits in suffix of splitted files')
    parser.add_argument('--splitsize', '-s', dest='splitSize', type=int, default=None, help='assign the size (number of lines) of each splitted file')
    parser.add_argument('--chunks', '-c', dest='numChunks', type=int, default=None, help='assign the number of splitted files (default: same as --threads parameter)')
    parser.add_argument('--threads', '-n', type=int, default=None, help='assign the maximum number of worker processes (default: %s in your computer' % numCPUs)
    #parser.add_argument('--tmpdir', '-t', type=str, default='./tmp', help='assign the path of working directory')
    parser.add_argument('--tmpdir', '-t', type=str, default=None, help='assign the path of working directory (default: "./tmp-[basename]"')
    parser.add_argument('--verbose', '-v', action='store_true', help='verbosely print progressive messages')
    parser.add_argument('--interval', '-i', type=float, default=SLEEP_DURATION, help='sleep time duration for each confirmation (default: %(default)s)')
    parsed = parser.parse_args(args)
    #conf = Config(vars(parsed))
    conf = Config()
    conf.update(vars(parsed))
    logger.debug(conf)
    #execParallel(**vars(parsed))
    execParallel(conf)

def main():
    cmdExecParallel(sys.argv[1:])

if __name__ == '__main__':
    main()




# distutils: language=c++
# -*- coding: utf-8 -*-

'''functions mapping from words/phrases to IDs and vice versa'''

# C++ setting
from libcpp cimport bool

# Local libraries
from lpu.data_structs.trie import TwoWayIDMap

#wordMap   = TwoWayIDMap()
#wordMap   = {}
phraseMap = TwoWayIDMap()
#phraseMap = {}

cdef class StringEnumerator:
    # defined in vocab.pxd
    # cdef dict dict_str2id
    # cdef list list_id2str

    def __cinit__(self):
        self.dict_str2id = {}
        self.list_id2str = []

    cpdef bool append(self, str string):
        self.str2id(string)
        return True

    cpdef long str2id(self, str string):
        cdef long new_id
        if string in self.dict_str2id:
            return self.dict_str2id[string]
        else:
            new_id = len(self.list_id2str)
            self.list_id2str.append(string)
            self.dict_str2id[string] = new_id
            return new_id

    cpdef str id2str(self, long number):
        if 0 <= number and number < len(self.list_id2str):
            return self.list_id2str[number]
        else:
            raise IndexError("id %s is not registered in vocabulary set" % (number,))

    def ids(self):
        cdef long i = 0, length = len(self.list_id2str)
        while i < length:
            yield i
            i += 1

    def strings(self):
        cdef str string
        for string in self.list_id2str:
            yield string

    def __iter__(self):
        return self.strings()

    def __len__(self):
        return len(self.list_id2str)

cdef StringEnumerator word_enum   = StringEnumerator()
cdef StringEnumerator phrase_enum = StringEnumerator()

cpdef long word2id(str word):
    return word_enum.str2id(word)
    #return wordMap[word]

cpdef str id2word(long number):
    return word_enum.id2str(number)
    #return wordMap.id2str(number)

cpdef str phrase2idvec(str phrase):
    if not phrase:
        return ''
    #return str.join(',', map(str, map(word2id, phrase.split(' '))))
    #return str.join(',', map(str, map(word2id, phrase.split())))
    return str.join(',', map(str, map(word2id, phrase.strip().split(' '))))

cpdef str idvec2phrase(str idvec):
    if not idvec:
        return ''
    return str.join(' ', map(id2word, map(int, idvec.split(','))))

cpdef long phrase2id(str phrase):
    cdef str idvec = str.join(',', map(str, map(word2id, phrase.split(' '))))
    return phraseMap[idvec]

cpdef str id2phrase(long number):
    cdef str idvec = phraseMap.id2str(number)
    return str.join(' ', map(id2word, map(int, idvec.split(','))))




#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''functions to normalize phrase translation probabilities of given phrase/rule table'''

# Standarde libraries
import argparse
import math
import sys

# Local libraries
from lpu.common import files
from lpu.common import logging
from lpu.common import progress
from lpu.smt.trans_models.records import MosesRecord, TravatarRecord
from lpu.smt.trans_models.tables import Table

#from lpu.data_structs import trees

def get_target_field(rec, target_number=None):
    if isinstance(target_number,int):
        return rec.trg.split('|COL|')[target_number].strip()
    else:
        return rec.trg

def calc_src_factor(table, src, target_number=None):
    total = 0
    feature = 'egfp'
    targets = set()
    if target_number:
        feature = str(target_number) + feature
    for i, rec in enumerate( table.find_src(src) ):
        target = get_target_field(rec,target_number)
        if target not in targets:
            total += rec.features[feature]
            targets.add(target)
    return total

def calc_trg_factor(table, trg, target_number=None):
    total = 0
    feature = 'fgep'
    if target_number:
        feature = str(target_number) + feature
    sources = set()
    #for i, rec in enumerate( table.find_trg(trg) ):
    #for i, rec in enumerate( table.find(trg) ):
    for i, rec in enumerate( table.find_trg(trg, target_number) ):
        #logging.log((i, rec.to_str()))
        #if target_number == 0:
        #    logging.log(trg)
        #    logging.log(get_target_field(rec,target_number))
        #    logging.log(rec.to_str())
        #    logging.log(feature)
        #    logging.log(rec.features[feature])
        #    logging.log("--")
        #if trg == get_target_field(rec,target_number):
        if rec.src not in sources:
            total += rec.features[feature]
            sources.add(rec.src)
    return total

def normalize_table(table_path, save_path):
    table = Table(table_path, TravatarRecord, trg_key=True)

    last_src = ""
    #src_total = 0
    src_total = {}
    dict_trg_total = {}
    with files.open(save_path, 'wt') as fobj_out:
        #for i, rec in enumerate( progress.view(table, 'normalizing', max_count = len(table)) ):
        for i, line in enumerate( progress.view(table_path, 'normalizing', max_count = len(table)) ):
            try:
                rec = table.RecordClass(line)
                #if i > 100:
                #    break
                targets = [rec.trg]
                target_numbers = [None]
                if rec.trg.find('|COL|') >= 0:
                    for i, field in enumerate(rec.trg.split('|COL|')):
                        targets.append(field.strip())
                        target_numbers.append(i)
                old = rec.to_str()
                if last_src != rec.src:
                    for num in target_numbers:
                        src_total[num] = calc_src_factor(table, rec.src, num)
                    last_src = rec.src
                for num, target in zip(target_numbers, targets):
                    #logging.log(target)
                    trg_track = table.field_dict.get_node(target).track()
                    if (num,trg_track) in dict_trg_total:
                        trg_total = dict_trg_total[num,trg_track]
                    else:
                        #trg_total = calc_trg_factor(table, rec.trg, num)
                        trg_total = calc_trg_factor(table, target, num)
                        dict_trg_total[num, trg_track] = trg_total
                    prefix = ''
                    if isinstance(num,int):
                        prefix=str(num)
                    rec.features[prefix+'egfp'] /= src_total[num]
                    rec.features[prefix+'fgep'] /= trg_total
                #if old != rec.to_str():
                #    logging.log("old: %s" % old)
                #    logging.log("new: %s" % rec.to_str())
                #    logging.log("--")
                fobj_out.write(rec.to_str())
                fobj_out.write("\n")
            except Exception as e:
                logging.warn("source: %s" % rec.src)
                logging.warn("target: %s" % rec.trg)
                logging.warn("num: %s" % num)
                logging.warn("target: %s" % target)
                logging.warn(e)
                raise Exception()

def main():
    parser = argparse.ArgumentParser(description = 'load 2 rule tables and pivot into one travatar rule table')
    parser.add_argument('table_path', type=str, help = 'path to phrase/rule table')
    parser.add_argument('save_path', help = 'path to save phrase/rule table')
    args = parser.parse_args()

    normalize_table(args.table_path, args.save_path)

if __name__ == '__main__':
    main()




# distutils: language=c++
# -*- coding: utf-8 -*-

'''functions for numbers'''

# Standard libraries
import math
import sys

from lpu.common import logging

logger = logging.getColorLogger(__name__)

cpdef object toNumber(anyNum, float margin = 0):
    cdef float floatNum
    #cdef int intNum
    cdef long intNum
    floatNum = float(anyNum)
    intNum   = int(round(floatNum))
    if abs(floatNum - intNum) <= margin:
        return intNum
    else:
        return floatNum

cpdef bytes __py2__intToBytes(n, int length, str byteorder='big'):
    assert n >= 0
    strHex = '%x' % n
    if len(strHex) >= length*2:
        raise OverflowError('int too big to convert')
    strDecoded = strHex.zfill(length*2).decode('hex')
    return strDecoded if byteorder == 'big' else strDecoded[::-1]
cpdef bytes __py3__intToBytes(n, int length, str byteorder='big'):
    return int.to_bytes(n, length, byteorder)

cpdef __py2__intFromBytes(bytes b, str byteorder='big'):
    if byteorder == 'big':
        return int(b.encode('hex'), 16)
    else:
        return int(b[::-1].encode('hex'), 16)
cpdef __py3__intFromBytes(bytes b, str byteorder='big'):
    return int.from_bytes(b, byteorder)

if sys.version_info.major == 2:
    # Python2
    intFromBytes = __py2__intFromBytes
    intToBytes   = __py2__intToBytes
elif sys.version_info.major == 3:
    # Python3
    intFromBytes = __py3__intFromBytes
    intToBytes   = __py3__intToBytes
else:
    raise SystemError("Unsupported python version: %s" % sys.version)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os.path
import sys

from setuptools import findall
from setuptools import find_packages
from setuptools import setup
from setuptools import Extension

import numpy

MAIN_PACKAGE = 'lpu'

if sys.platform == 'linux2':
    from distutils import sysconfig
    # Omitting 'strict-prototypes' warning For Python 2.x
    opt = sysconfig.get_config_vars().get('OPT', '')
    opt = opt.replace('-Wstrict-prototypes', '')
    sysconfig.get_config_vars()['OPT'] = opt
    # Omitting 'strict-prototypes' warning For Python 3.x
    cflags = sysconfig.get_config_vars().get('CFLAGS', '')
    cflags = cflags.replace('-Wstrict-prototypes', '')
    sysconfig.get_config_vars()['CFLAGS'] = cflags

def get_package(path):
    if os.path.exists( os.path.join(path, '__init__.py') ):
        dirpath = os.path.dirname( os.path.abspath(str(path)) )
        parent = get_package(dirpath)
        if parent:
            return parent + '.' + os.path.basename(path)
        return os.path.basename(path)
    return ''

def get_modules(dirpath, suffixes=['.pyx'], additional_sources=None, **options):
    modules = []
    for path in findall(dirpath):
        if isinstance(suffixes, str):
            suffixes = [suffixes]
        if not any([path.endswith(suffix) for suffix in suffixes]):
            continue
        dirpath = os.path.dirname(path)
        module = os.path.splitext( os.path.basename(path) )[0]
        package = get_package(dirpath)
        if package:
            module = package + '.' + module
        sources = [path]
        if additional_sources:
            sources = sources + list(additional_sources)
        modules.append( Extension(module, sources, **options) )
    return modules

include_dirs = [
    numpy.get_include(),
]

try:
    from Cython.Build import build_ext
    from Cython.Build import cythonize
    compiler_directives = dict(
        language_level = sys.version_info[0],
    )
    ext_modules = get_modules(
        MAIN_PACKAGE, ['.pyx'],
        include_dirs = include_dirs,
    )
    ext_modules = cythonize(
        ext_modules,
        compiler_directives = compiler_directives,
    )
    cmdclass = {'build_ext': build_ext}
except ImportError:
    ext_modules = get_modules(
        MAIN_PACKAGE, ['.c', '.cpp'],
        include_dirs = include_dirs,
    )
    cmdclass = {}

version_file = os.path.join(os.path.dirname(__file__), MAIN_PACKAGE, 'VERSION')
version = open(version_file).read().strip()

install_requires = [
    #'Cython',
    'numpy',
]

setup(
    name = 'lpu',
    version = version,
    cmdclass = cmdclass,
    ext_modules = ext_modules,
    install_requires = install_requires,
    description = 'A Language Processing Utility',
    long_description = open('README.md').read(),
    long_description_content_type = 'text/markdown',
    url = 'https://github.com/akivajp/lpu',
    author = 'Akiva Miura',
    author_email = 'akiva.miura@gmail.com',
    license = 'MIT',
    keywords = [
        'CL',
        'NLP',
        'computational linguistics',
        'natural language processing',
    ],
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Utilities",
    ],
    packages = find_packages(),
    package_data = {
        '': [
            '*.pxd',
            '*.pyx',
        ],
        'lpu': [
            'VERSION'
        ],
    },
    entry_points = {
        'console_scripts': [
            'lpu-abspath = lpu.commands.abspath:main',
            'lpu-clean-parallel= lpu.commands.clean_parallel:main',
            'lpu-dialog  = lpu.commands.prompt:main',
            'lpu-exec-parallel= lpu.commands.exec_parallel:main',
            'lpu-guess-langcode= lpu.commands.guess_langcode:main',
            'lpu-progress= lpu.commands.progress:main',
            'lpu-random-split= lpu.commands.random_split:main',
            'lpu-train-ibm-model1= lpu.smt.align.ibm_model1:main_train',
            #'lpu-score-ibm-model1= lpu.smt.align.ibm_model1:main_score',
            'lpu-wait-files= lpu.commands.wait_files:main',
        ],
    },
)

